#!/usr/bin/env python3
"""Create a data transport application"""

##########################################################################
#
#   Create initial data transport application directory layout.
#
#   2020-10-09  Todd Valentic
#               Initial implementation
#
##########################################################################

import argparse
import grp
import logging
import os
import platform
import pwd
import shutil
import sys

from pathlib import Path

import datatransport

if sys.version_info < (3, 10):
    import importlib_resources as resources
else:
    from importlib import resources

def parse_args(args=None):
    """Parse command line"""

    if not args:
        args = sys.argv[1:]

    user = pwd.getpwuid(os.getuid()).pw_name
    group = grp.getgrgid(os.getgid()).gr_name
    hostname = platform.node().split(".", 1)[0]

    parser = argparse.ArgumentParser(
        description="Create a new data transport application"
    )

    parser.add_argument("-d", "--dir", help="Installation directory", default=".")

    parser.add_argument("-u", "--user", help=f'Run as user "{user}"', default=user)

    parser.add_argument("-g", "--group", help=f'Run as group "{group}"', default=group)

    parser.add_argument(
        "-n", "--hostname", help=f'Hostname "{hostname}"', default=hostname
    )

    parser.add_argument("-v", "--verbose", help="Verbose output", action="store_true")
    return parser.parse_args(args)


def create_dirs(basepath):
    """Create directory layout"""

    basepath.mkdir(exist_ok=True, parents=True)

    dirs = ["bin", "etc", "var", "tmp", "log", "groups"]

    for leaf in dirs:
        basepath.joinpath(leaf).mkdir(exist_ok=True)

    basepath.joinpath("groups").chmod(0o2775)


def install_config(args, basepath):
    """Install configuration files"""

    for entry in resources.files("datatransport.templates.etc").iterdir():
        contents = entry.read_text()
        contents = contents.replace("@PREFIX@", str(basepath.absolute()))
        contents = contents.replace("@PACKAGE_VERSION@", datatransport.__version__)
        contents = contents.replace("@USER@", args.user)
        contents = contents.replace("@GROUP@", args.group)
        contents = contents.replace("@HOSTNAME@", args.hostname)

        output_path = basepath / "etc" / entry.name

        with output_path.open(mode="w") as output:
            output.write(contents)


def link_programs(basepath):
    """Link programs into bin"""

    venv_path = Path(os.environ["VIRTUAL_ENV"], "bin")
    dest_path = basepath.joinpath("bin")

    programs = [
        "cancelnewsgroup",
        "console",
        "createnewsgroup",
        "listnewsgroups",
        "rmnewsgroup",
        "transport-create-app",
        "transport-get-article",
        "transport-post-article",
        "transportctl",
        "transportd",
        "transportps",
        "viewlog",
    ]

    for program in programs:
        src = venv_path.joinpath(program)
        dst = dest_path.joinpath(program)
        dst.symlink_to(src)


def install_groups(package, basepath):
    """Install default process groups"""

    basepath.mkdir(exist_ok=True)
    basepath.chmod(0o2775)

    for entry in [r.name for r in resources.files(package).iterdir()]:
        if resources.files(package).joinpath(entry).is_file():
            with resources.path(package, entry) as srcpath:
                destpath = basepath / entry
                shutil.copyfile(srcpath, destpath)
                destpath.chmod(0o775)
                logging.debug("Copy %s -> %s", srcpath, basepath)
        else:
            install_groups(f"{package}.{entry}", basepath / entry)


def main():
    """Script entry point"""

    args = parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    basepath = Path(args.dir)

    create_dirs(basepath)
    install_config(args, basepath)
    link_programs(basepath)
    install_groups("datatransport.templates.groups", basepath / "groups")

    print(f"Created in {basepath}")



#!/usr/bin/env python3
"""Post files to a newsgroup"""

#####################################################################
#
#   Post a files in a message to a newsgroup
#
#   2016-12-26  Todd Valentic
#               Use datatransport package
#
#   2023-08-23  Todd Valentic
#               Updated for transport3 / python3
#
#####################################################################

import argparse
import pathlib
import sys

from datatransport import newstool

VERSION = "1.0"


def main():
    """Script entry point"""

    desc = "Remove a newsgroup from the news server"
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument("-V", "--version", action="version", version=VERSION)

    parser.add_argument(
        "-s", "--server", default="localhost", help="News server (default: localhost)"
    )
    parser.add_argument(
        "-p", "--port", default=119, type=int, help="News server port (default: 119)"
    )
    parser.add_argument(
        "-t", "--text", action="store_true", help="Post file contents as body text"
    )
    parser.add_argument("-n", "--newsgroup", help="Newsgroup name", required=True)
    parser.add_argument("-b", "--subject", help="Subject header")
    parser.add_argument("filenames", nargs="+", metavar="filename", type=pathlib.Path)

    args = parser.parse_args()

    poster = newstool.NewsPoster()
    poster.set_server(args.server, port=args.port)
    poster.set_newsgroup(args.newsgroup)
    if args.subject:
        poster.set_subject(args.subject)

    if not poster.has_newsgroup(args.newsgroup):
        print(f"The newsgroup does not exist: {args.newsgroup}")
        sys.exit(1)

    if args.text:
        content = args.filenames[0].read_text("utf-8")
        poster.post_text(content)
    else:
        poster.post(args.filenames)

    sys.exit(0)



##############################################################################
#
#  NOTICE: Deprectated. Use NewsgroupMonitor for new projects.
#
#  InstrumentStatus
#
#  This script polls newsgroups to determine if an instrument is running.
#  The output is an HTML fragment that can be included in web pages.
#
#  There is a plugin callback system available to chain into the events
#  for each instrument. You need to create a python module with a callable
#  function and make sure that it is with in the python path. If you have
#  the plugin in the group's home, then you will want to make sure to add
#  this to the PYTHONPATH. Add the following line in the ProcessGroup
#  section of the config file:
#
#  environ.add.pythonpath: %(group.home)s
#
#  The function must be callable with the following keyword parameters:
#
#      callback(self,curstate=None,prevstate=None,changed=0)
#
#  When the module is first loaded, the init method will be called.
#  You can initialize any variables that you need
#  by storing then into self.*. You can even pull in values from the
#  config file (which is cool!). During the run of the program, the
#  callback will be called with a code in the even field. The current
#  codes are defined in Status in this module. If the state has
#  changed since the last check, then the changed flag is set.
#
#  History:
#
#   1.0.0   2000-02-22  TAV
#           Initial implementation
#
#   1.0.1   2000-02-22  TAV
#           Added maillist reporting.
#
#   1.0.2   2000-02-26  TAV
#           Added ability to set the table title from the config file.
#
#   1.0.3   2000-02-29  TAV
#           Updated appearance of HTML tables (colors and fonts).
#
#   1.0.4   2000-03-14  TAV
#           Another HTML update - now uses simplified table and style sheets.
#
#   1.0.5   2000-04-26  TAV
#           Added the reportFailure parameter to the instruments. This will
#               cause a mail message to be sent to the troublegroup news
#               group. The intention is to notify people when an instrument
#               that should always be running has stopped.
#
#   1.0.6   2000-10-26  TAV
#           Added the reportStart parameter to complement the reportStop
#               feature. This change allows system status to signal when
#               an instrument such as the imager or lidar begin sending data.
#
#   1.0.7   2001-11-09  TAV
#           Renamed newstool -> NewsTool
#
#   1.0.8   2001-12-02  TAV
#           Fixed problem connecting to newsserver due to NewsTool name
#               change. Also added some log messages to indicate alert
#               to problem in future.
#           Cleaned up some HTML.
#
#   1.0.9   2002-09-03  Todd Valentic
#           Moved to python2, sri.transport -> Transport
#
#   1.0.10  2002-10-04  Todd Valentic
#           Fixed some config file name bugs.
#           Added traceback listing if count get operation fails.
#
#   1.0.11  2002-11-10  Todd Valentic
#           Added more tracebacks.
#           Check for self.running, use self.wait()
#
#   1.0.12  2003-02-21  Todd Valentic
#           Improved determination of time from header
#
#   1.0.13  2003-03-26  Todd Valentic
#           Assumptions were being made that the time was GMT. Now using
#               ISO format for printing in messages.
#           Using DateTime module versus time for everything.
#           Changed the time determination again. Now I'm just using the
#               NNTP-Posting-Date header instead of trying to parse the
#               header. This header reflects the time that the last message
#               was posted into the transport network.
#           Removed string module (using methods instead).
#
#   1.0.14  2003-04-04  Todd Valentic
#           Changed the pollserver and pollrate config parameters to be
#               poll.newsserver and poll.rate so they are consistent with
#               other naming schemes.
#           Fixed newly instroduced bug in instrument check time.
#           Changed reportStart and reportStop to report.start and report.stop
#
#   1.1.0   2003-04-07  Todd Valentic
#           Started major overhaul. Now remember state between runs.
#
#   1.1.1   2003-04-18  Todd Valentic
#           Changed window -> timeout for config parameter.
#           Default the timeout to 5 minutes.
#
#   1.1.2   2003-05-11  Todd Valentic
#           Added link to parent.abort in Instrument class.
#           Added plugin system to hook into events.
#
#   1.1.3   2003-06-11  Todd Valentic
#           Converted plugin to new general callback scheme.
#
#   1.1.4   2003-08-14  Todd Valentic
#           Problem using the general callback scheme with Instrument.
#               need to make some sort of mixin class.
#
#   1.1.5   2004-08-28  Todd Valentic
#           Updated to XML-RPC version.
#           Import PageKit from Componentsa
#
#   1.1.6   2004-12-13  Todd Valentic
#           Removed traceback call (left over from XML-RPC conversion)
#           Renamed self.config -> self.history to prevent name collision.
#
#   1.1.7   2004-12-27  Todd Valentic
#           Convert from mx.DateTime to datetime
#
#   1.1.8   2005-01-26  Todd Valentic
#           Added get_boolean to Instrument to track change 
#
#   1.1.9   2005-05-04  Todd Valentic
#           Added getfloat() to instrumnent (bug fix)
#
#   1.1.10  2005-07-07  Todd Valentic
#           Use get_timedelta for poll.rate
#
#   1.1.11  2007-08-06  Todd Valentic
#           Added history posting
#
#   1.1.12  2007-08-08  Todd Valentic
#           Fixed typo in Instrument news server parameter
#
#   1.1.13  2008-11-13  Todd Valentic
#           Use SafeConfigParser
#
#   1.1.14  2008-12-11  Todd Valentic
#           Make sure only strings are used in config
#
#   1.1.15  2009-05-05  Todd Valentic
#           Fixed bug in log message when posting failed.
#
#   2016-12-23  Todd Valentic
#               Use datetime.timezone for utc
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#               Python3 migration
#                   ConfigParser -> sapphire.Parser 
#                   StringIO -> io.StringIO
#                   NewsPoster
#
##############################################################################

import datetime
import io
import sys

from dateutil import parser

from datatransport import ProcessClient
from datatransport import NewsPoster
from datatransport import newstool
from datatransport import ConfigComponent
from datatransport.apps import PageKit

import sapphire_config as sapphire

##############################################################################


class InstrumentTable(PageKit.Table):
    def __init__(self, title, instruments):
        PageKit.Table.__init__(self, title=title)

        self.instruments = instruments
        self.num_cols = 2

    def print_status(self, instrument):

        print("<tr>")
        print('<td class=desc width="85%">', end="")
        print("&nbsp;&nbsp;%s" % instrument.label, end="")
        print("</td>")

        if instrument.status == Status.online:
            msg = "Running"
            style = "running"
        elif instrument.status == Status.offline:
            msg = "Stopped"
            style = "stopped"
        else:
            msg = "Unknown"
            style = "warning"

        print("<td class=%s>%s</td>" % (style, msg))
        print("</tr>")

    def body(self):

        for instrument in self.instruments:
            self.print_status(instrument)


##############################################################################


class Status:

    unknown = 0
    offline = 2
    online = 3

    label = {unknown: "unknown", offline: "offline", online: "online"}


##############################################################################


class Instrument(ConfigComponent):
    def __init__(self, name, parent):
        ConfigComponent.__init__(self, None, name, parent)

        self.news_poster = NewsPoster(self, quiet=1)

        self.load(parent.history)

        self.label = self.get("label", name)
        self.pollserver = self.get("poll.newsserver", "localhost")
        self.pollgroup = self.get("poll.newsgroup", "")
        self.reportstart = self.get_boolean("report.start", True)
        self.reportstop = self.get_boolean("report.stop", True)
        self.timeout = self.get_timedelta("timeout", 300)
        self.callback = self.get_callback("callback")

        if self.callback:
            self.callback(self, curstatus=self.status)

        self.newstool = newstool.NewsTool()
        self.newstool.set_server(self.pollserver)
        self.newstool.set_newsgroup(self.pollgroup)

        self.log.info("  " + name)
        self.log.info("    label       : %s" % self.label)
        self.log.info("    poll server : %s" % self.pollserver)
        self.log.info("    poll group  : %s" % self.pollgroup)
        self.log.info("    timeout     : %s" % self.timeout)
        self.log.info("    report start: %d" % self.reportstart)
        self.log.info("    report stop : %d" % self.reportstop)
        self.log.info("    status      : %s" % Status.label[self.status])
        self.log.info("    start time  : %s" % self.starttime)
        self.log.info("    stop time   : %s" % self.stoptime)
        self.log.info("    first msg   : %s" % self.firstmessage)
        self.log.info("    last msg    : %s" % self.lastmessage)
        self.log.info("    run time    : %s" % self.runtime)
        self.log.info("    off time    : %s" % self.offtime)

        module = self.get("callback.module")
        function = self.get("callback.function")

        if self.callback:
            self.log.info("    callback    : %s.%s" % (module, function))
        else:
            self.log.info("    callback    : None")

    def save(self, config):

        if not config.has_section(self.name):
            config.add_section(self.name)

        config.set(self.name, "status", str(self.status))
        config.set(self.name, "starttime", str(self.starttime))
        config.set(self.name, "stoptime", str(self.stoptime))
        config.set(self.name, "firstmessage", str(self.firstmessage))
        config.set(self.name, "lastmessage", str(self.lastmessage))
        config.set(self.name, "runtime", str(self.runtime.seconds))
        config.set(self.name, "offtime", str(self.offtime.seconds))
        config.set(self.name, "label", self.label)

    def loadtime(self, config, key):

        try:
            return parser.parse(config.get(self.name, key))
        except:
            return datetime.datetime.now(datetime.timezone.utc)

    def loaddeltatime(self, config, key):

        try:
            secs = config.get_float(self.name, key)
            return datetime.timedelta(seconds=secs)
        except:
            return datetime.timedelta(0)

    def load(self, config):

        try:
            self.status = config.get_int(self.name, "status")
        except:
            self.status = Status.unknown

        self.starttime = self.loadtime(config, "starttime")
        self.stoptime = self.loadtime(config, "stoptime")
        self.firstmessage = self.loadtime(config, "firstmessage")
        self.lastmessage = self.loadtime(config, "lastmessage")

        self.runtime = self.loaddeltatime(config, "runtime")
        self.offtime = self.loaddeltatime(config, "offtime")

    def check(self):

        prevstatus = self.status
        self.status = self.get_status()

        self.changed = 0

        # unknown -> online

        if prevstatus == Status.unknown and self.status == Status.online:

            self.starttime = self.firstmessage

            self.log.info("  %s state now known (online)" % self.name)

        # unknown -> offline

        elif prevstatus == Status.unknown and self.status == Status.offline:

            self.stoptime = self.servertime
            self.starttime = self.servertime

            self.log.info("  %s state now known (offline)" % self.name)

        # unknown -> unknown

        elif prevstatus == Status.unknown and self.status == Status.unknown:

            self.log.debug("  %s state still unknown" % self.name)

        # offline -> online

        elif prevstatus == Status.offline and self.status == Status.online:

            self.changed = 1
            self.starttime = self.firstmessage
            self.changetime = self.starttime
            self.offtime = self.starttime - self.stoptime

            if self.reportstart:
                header, message = self.start_message()
                self.post(header, message)

            self.log.info("  %s going online (was off %s)" % (self.name, self.offtime))

        # online -> offline

        elif prevstatus == Status.online and self.status == Status.offline:

            self.changed = 1
            self.stoptime = self.lastmessage
            self.changetime = self.stoptime
            self.runtime = self.stoptime - self.starttime

            if self.reportstop:
                header, message = self.stop_message()
                self.post(header, message)

            self.log.info("  %s going offline (was on %s)" % (self.name, self.runtime))

        # still online

        elif prevstatus == Status.online and prevstatus == self.status:

            self.stoptime = self.lastmessage
            self.runtime = self.stoptime - self.starttime

            self.log.debug("  %s still online (%s)" % (self.name, self.runtime))

        # still offline

        elif prevstatus == Status.offline and prevstatus == self.status:

            self.offtime = self.servertime - self.stoptime

            self.log.debug("  %s still offline (%s)" % (self.name, self.offtime))

        if self.callback:
            try:
                self.callback(
                    self,
                    curstatus=self.status,
                    prevstatus=prevstatus,
                    changed=self.changed,
                )
            except:
                self.log.exception("Problem running the callback function:")

        return self.changed

    def get_message_time(self, msg):

        # Date is UTC

        try:
            server = self.newstool.open_server()
            response = server.group(self.pollgroup)
            date = server.xhdr("NNTP-Posting-Date", msg)[1][0]
            date = " ".join(date.split()[1:])
            date = parser.parse(date)
            server.quit()
        except:
            date = datetime.datetime.now(datetime.timezone.utc)
            self.log.exception("Problem determining message time")

        return date

    def get_status(self):

        articles = self.newstool.list_articles(self.timeout).values()[0]

        self.servertime = self.newstool.datetime()

        if len(articles) > 0:

            self.firstmessage = self.get_message_time(articles[0])
            self.lastmessage = self.get_message_time(articles[-1])

            return Status.online

        else:

            return Status.offline

    def start_message(self):

        header = "%s has started" % self.label

        message = []
        message.append("")
        message.append("  The following instrument has started sending data:")
        message.append("")
        message.append("    Instrument: %s" % self.label)
        message.append("    Date/Time : %s" % self.starttime)
        message.append("    Offline   : %s" % self.offtime)
        message.append("")

        message = "\n".join(message)

        return header, message

    def stop_message(self):

        header = "%s has stopped" % self.label

        message = []
        message.append("")
        message.append("  The following instrument has stopped sending data:")
        message.append("")
        message.append("    Instrument: %s" % self.label)
        message.append("    Date/Time : %s" % self.stoptime)
        message.append("    Online    : %s" % self.runtime)
        message.append("")

        message = "\n".join(message)

        return header, message

    def post(self, header, message):

        try:
            self.news_poster.set_subject(header)
            self.news_poster.post_text(message)
        except:
            self.log.exception("Error posting message for %s" % self.name)


##############################################################################


class InstrumentStatus(ProcessClient):
    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.news_poster = NewsPoster(self)
        self.pollrate = self.get_timedelta("poll.rate", 60)
        self.htmlname = self.get("htmlname", "instrumentstatus.html")
        self.title = self.get("title", "Instrument Status")
        self.history = sapphire.Parser()

        self.history_poster = self.create_news_poster("history", quiet=True)

        self.load_history()

        self.log.info("Reading instrument definitions:")

        self.instruments = []
        self.labelwidth = 0

        for name in self.get("instruments", "").split():
            try:
                instrument = Instrument(name, self)
                self.instruments.append(instrument)
                self.labelwidth = max(self.labelwidth, len(instrument.label))
            except:
                self.log.exception("Problem creating %s" % name)
                self.abort("Exiting")

        self.save_history()

    def load_history(self):

        self.log.info("Loading history")

        try:
            self.history.read("history")
        except:
            self.log.exception("Problem reading history file:")
            self.abort("Exiting")

    def save_history(self):

        self.history = sapphire.Parser()

        self.log.debug("Saving history")

        for instrument in self.instruments:
            instrument.save(self.history)

        self.history.write(open("history", "w"))

        buffer = io.StringIO.StringIO()
        self.history.write(buffer)

        try:
            self.history_poster.post_text(buffer.getvalue())
        except:
            self.log.error("Failed to post history")

    def post_message(self):

        message = []
        message.append("")
        message.append("  A change in instrument status has been detected:")
        message.append("")

        header = None

        for instrument in self.instruments:

            status = Status.label[instrument.status].capitalize()

            if instrument.changed == 0:
                changed = ""
            else:
                changed = "<-- %s" % instrument.changetime
                if not header:
                    header = instrument.label + " " + status
                else:
                    header = "Multiple instruments have changed status"

            instrument.changed = 0

            message.append(
                "  %s: %-10s %s"
                % (instrument.label.rjust(self.labelwidth), status, changed)
            )

        message = "\n".join(message)

        try:
            self.news_poster.set_subject(header)
            self.news_poster.post_text(message)
        except:
            self.log.error("Error posting message to mailing list.")

    def report(self):

        InstrumentTable(self.title, self.instruments).save(self.htmlname)

        self.post_message()

        self.log.info("Updated status report")

    def instrument_change(self):

        self.log.debug("Checking for status change:")

        numchanged = 0

        for instrument in self.instruments:

            try:
                instrument.check()
            except:
                self.log.exception("Problem checking %s" % instrument.name)
                continue

            numchanged += instrument.changed

            if self.is_stopped():   
                break

        return numchanged > 0

    def main(self):

        while self.wait(self.pollrate):

            if self.instrument_change():
                self.save_history()
                self.report()

        self.save_history()


def main():
    InstrumentStatus(sys.argv).run()



#!/usr/bin/env python

#####################################################
#
#   Genset Plotter
#
#   Extract timestamp info from DEFAULT section
#
#   2010-08-21  Todd Valentic
#               Initial implementation
#
#####################################################

from Transport.Components   import PlotTool
from Transport.Util         import removeFile
from Transport              import NewsTool

from dateutil import parser

import sys
import ConfigParser

class PlotGenset(PlotTool):

    def parse(self,message):

        filenames = NewsTool.saveFiles(message)

        self.data = ConfigParser.ConfigParser()
        self.data.read(filenames)

        removeFile(filenames)

        return True

    def getTimestamp(self,message):
        timestamp = self.data.get('DEFAULT','timestamp')
        timestamp = parser.parse(timestamp).replace(tzinfo=self.utc)
        return timestamp

if __name__ == '__main__':
    PlotGenset(sys.argv).run()



#!/usr/bin/env python3
"""Newsgroup Gateway Component"""

#########################################################################
#
#   News group gateway service
#
#   This service provides an XML-RPC gateway to news groups.
#   The contents of the latest available message in a group
#   are cached and made available to client programs.
#
#   2007-04-07  Todd Valentic
#               Initial implementation.
#
#   2007-12-06	Todd Valentic
#               Make read() return a JSON representation of the message.
#
#   2009-11-24  Todd Valentic
#               Use currentTime instead of datetime.now()
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#               NewsPoller
#
#   2023-07-26  Todd Valentic
#               Updated for transport3 / python3
#
#########################################################################

import pickle
import shelve
import sys

from threading import Thread, Lock

from datatransport import ProcessClient
from datatransport import XMLRPCServer
from datatransport import NewsPoller
from datatransport import AccessMixin


def synchronized(lock):
    """Lock decorator"""

    def wrap(f):
        def new_function(*args, **kw):
            lock.acquire()
            try:
                return f(*args, **kw)
            finally:
                lock.release()

        return new_function

    return wrap


class Data(AccessMixin):
    """Data cache"""

    lock = Lock()

    def __init__(self, parent):
        AccessMixin.__init__(self, parent)
        self.messages = shelve.open("cache")

    @synchronized(lock)
    def write(self, message):
        """Store message into cache"""

        newsgroup = message["Newsgroups"]
        self.messages[newsgroup] = self.write_handler(newsgroup, message)
        self.messages.sync()

    def write_handler(self, _newsgroup, message):
        """What to cache"""
        return message

    @synchronized(lock)
    def clear(self):
        """Clear the cache"""

        self.messages.clear()
        self.messages.sync()

    @synchronized(lock)
    def read(self, newsgroup):
        """Read entry from cache"""

        return self.messages[newsgroup]

    @synchronized(lock)
    def list(self):
        """List cache"""

        return list(self.messages.keys())

    @synchronized(lock)
    def get_string(self, newsgroup):
        """Return message as string"""

        return self.messages[newsgroup].as_string()

    @synchronized(lock)
    def unpickle(self, newsgroup):
        """Unpickle message"""

        payload = self.messages[newsgroup].get_payload().encode("utf-8")
        return pickle.loads(payload)


class Server(Thread, AccessMixin):
    """XMLRPC Server Interface"""

    def __init__(self, parent, data):
        Thread.__init__(self)
        AccessMixin.__init__(self, parent)

        self.xmlserver = XMLRPCServer(self)
        self.run = self.xmlserver.main

        self.daemon = True

        self.data = data

        self.xmlserver.register_function(self.read, "read")
        self.xmlserver.register_function(self.data.unpickle, "unpickle")
        self.xmlserver.register_function(self.data.list, "list")
        self.xmlserver.register_function(self.data.get_string, "message")

        self.add_handlers()

        self.log.info("*** init done")

    def add_handlers(self):
        """Used by derived classes to add handlers"""
        return

    def read(self, newsgroup):
        """Read message from newsgroup"""

        message = self.data.read(newsgroup)
        headers = dict(message.items())
        return {"headers": headers, "body": message.get_payload()}


class NewsGateway(ProcessClient):
    """Process Client"""

    def __init__(self, dataFactory, serverFactory, argv):
        ProcessClient.__init__(self, argv)

        self.news_poller = NewsPoller(self, callback=self.process, idle=self.idle)
        self.main = self.news_poller.main

        self.data = dataFactory(self)
        self.timeout = self.config.get_timedelta("timeout")
        self.in_timeout = False

        if self.timeout:
            self.log.info("Timeout set to %s", self.timeout)
            self.deadline = self.now() + self.timeout
        else:
            self.log.info("No timeout set")

        serverFactory(self, self.data).start()

    def process(self, message):
        """Process handler"""

        self.log.info("Set from %s", message["Newsgroups"])
        self.data.write(message)

        if self.timeout:
            self.deadline = self.now() + self.timeout
            self.in_timeout = False

    def idle(self):
        """Idle handler"""

        if self.in_timeout:
            return

        if self.timeout and self.now() > self.deadline:
            self.log.info("timeout reached, clearing values")
            self.data.clear()
            self.in_timeout = True


def main():
    """Script entry point"""
    NewsGateway(Data, Server, sys.argv).run()



#!/usr/bin/env python
"""Scheduler Component"""

#############################################################################
#
#   Scheduler Process Client
#
#   Execute events at regular intervals. The schedule is defined in
#   the configuration file. Events are listed, one per line, in the
#   'events' parameter. Each line has a start time, duration, and
#   event name, separated by a the '|' character:
#
#   events:     00:00   |   00:45       | call-remote-1
#               05:00   |   00:45       | call-remote-2
#
#   The times are relative (timedelta) from the start of the event
#   sequence. The entire sequence is repeated at a rate defined by the
#   repear.* parameters.
#
#   The individual events are listed as 'event.<name>.<parameter>' The
#   specific parameters are dependent on the event types. Events are
#   objects that inherit from the Event class. An application would
#   provide it's own Event classes, passed in as the EventFactory
#   object to the Schedule init. Event object must support three
#   methods:
#
#       start   - called once at program startup
#       run     - called each time the event is scheduled
#       stop    - called once on exit
#
#   2005-04-07  Todd Valentic
#               Initial implementation.
#
#   2005-04-19  Todd Valentic
#               Added ability to start in mid-event sequence
#
#   2005-06-08  Todd Valentic
#               Added ability to start events early. This is useful
#                   in cases where multiple events are stacked up
#                   and some take less then there alloted time. It
#                   gives the extra time to the next event. Each
#                   event can be individually controlled, so those
#                   that need to happen at exact times still can.
#                   Even though an event may start early, it still
#                   runs until it's original end time.
#
#   2007-07-14  Todd Valentic
#               Computed offset in the wrong direction. Fixed.
#
#   2008-07-25  Todd Valentic
#               Add small offset in waituntil() to prevent problems
#                   due to small rounding errors.
#
#   2008-08-15  Todd Valentic
#               Still having problems with waituntil(). Now using
#                   a more robust scheme that will continue to
#                   retry/sleep until the target time is reached.
#
#   2009-11-24  Todd Valentic
#               Use currentTime()
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#
#   2023-07-26  Todd Valentic
#               Updated for transport3 / python3
#
#############################################################################

import datetime
import importlib
import sys

from datatransport import ProcessClient
from datatransport import ConfigComponent
from datatransport.utilities import datefunc


class SchedulerEvent(ConfigComponent):
    """Event Component"""

    def __init__(self, *p, **kw):
        ConfigComponent.__init__(self, "event", *p, **kw)
        self.startearly = self.config.get_boolean("startearly", False)

    def startup(self):
        """Start handler"""

    def shutdown(self):
        """Stop handler"""

    def run(self, endtime):
        """Run handler"""


class Scheduler(ProcessClient):
    """Scheduler Process Client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.rate = self.config.get_rate("repeat", 60)
        self.exit_on_error = self.config.get_boolean("exit_on_error", False)

        self.events = self.config.get_components("events", factory=self.event_factory)
        self.schedule = []

        lines = self.config.get_list("schedule", sep="\n")

        for lineno, entry in enumerate(lines):
            try:
                start, duration, name = entry.split("|")
                start = datefunc.parse_timedelta(start)
                duration = datefunc.parse_timedelta(duration)
                name = name.strip()
            except ValueError as err:
                self.log.error("Failed to parse schedule at line %d", lineno + 1)
                self.log.error(entry)
                self.log.error(str(err))
                self.abort()

            try:
                self.schedule.append((start, duration, self.events[name]))
            except KeyError:
                self.abort("Unkown event: %s", name)

        self.log.info("Repeat: %s", self.rate)

        self.log.info("Schedule:")

        for start, duration, name in self.schedule:
            self.log.info("  %s: %s - %s", name, start, start + duration)

    def event_factory(self, name, config, parent):
        """Create an event component"""

        key = config.get(f"event.{name}.type")
        key = config.get("event.default.type", key)
        key = config.get("event.*.type", key)

        module_name, method_name = key.rsplit(".", 1)

        module = importlib.import_module(module_name)

        return getattr(module, method_name)(name, config, parent)

    def event_cycle(self):
        """Single cycle throuth events in a schedule"""

        period = self.rate.period.total_seconds()

        for start_offset, duration, event in self.schedule:
            now = self.now()
            cycle_offset = datetime.timedelta(seconds=now.timestamp() % period)
            cycle_start = now - cycle_offset
            end_offset = start_offset + duration

            self.log.debug("Next event: %s", event.name)
            self.log.debug("cycle time: %s", cycle_offset)
            self.log.debug("start time: %s", start_offset)
            self.log.debug("end   time: %s", end_offset)

            if not event.startearly and cycle_offset < start_offset:
                self.log.debug("  waiting")
                if not self.wait(start_offset - cycle_offset):
                    return False

            self.log.info("Processing %s", event.name)

            if cycle_offset >= end_offset:
                self.log.info("  skipping, past end time")
                continue

            try:
                event.run(cycle_start + end_offset)
            except Exception as err:    # pylint: disable=broad-exception-caught
                if self.exit_on_error:
                    self.log.exception("Problem running event")
                    return False
                self.log.error("Problem running event: %s", err)

            self.log.info("  finished action for %s", event.name)

        return True

    def main(self):
        """Main application"""

        for event in self.events.values():
            event.startup()

        while self.wait(self.rate):
            self.log.info("---- Starting event cycle ----")

            if not self.event_cycle():
                break

            self.log.info("=" * 35)

        for event in self.events.values():
            event.shutdown()


def main():
    """Script entry point"""
    Scheduler(sys.argv).run()



#!/usr/bin/env python

#############################################################################
#
# Name:     PageKit
# Author:   Todd Valentic
# Company:  SRI International
#
# Purpose:  Base HTML page class for the transport network's dynamically
#           created pages.
#
# History:
#
#   1.0.0   1999-10-06  TAV
#           Initial implementation.
#
#   2.0.0   2000-03-13  TAV
#           Major cleanup and simplification (TransportPage disappeared).
#
#   2.0.1   2000-06-16  TAV
#           Added ability to delete as user.
#
#   2.0.2   2000-07-03  TAV
#           Added Boulder and Platteville to site name list (this *really*
#               needs to be fixed up and extracted outside of this file).
#
#   2.0.3   2000-08-15  TAV
#           Added Summitcamp to site name list. When am I going to fix this?
#
#   2.0.4   2001-05-02  TAV
#           Added SouthPole to site name list.
#
#   2.0.5   2001-10-23  TAV
#           Added the concept of user "roles" to generalize the admin
#               privledges, similar to a unix user's groups.
#
#   2.0.6   2001-11-09  TAV
#           Added InfoMessage class.
#
#   2.0.7   2001-12-26  TAV
#           Changed tav -> sri
#
#   2.0.8   2002-08-28  Todd Valentic
#           sri.transport -> Transport
#
#   3.0.0   2002-08-28  Todd Valentic
#           Reduced to basic layout tool.
#           Removed user features.
#
#   3.2.0   2002-11-11  Todd Valentic
#           Updated to modern stylesheet format.
#           Major code cleanup.
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#               Python3 updates
#                   print()
#
#############################################################################

import sys

from datatransport import TransportConfig


class Navbar:
    def __init__(self):

        self.choices = []
        self.title = "Navigation"

    def add_choice(self, label, url):
        self.choices.append((label, url))

    def show(self):

        print("<div id=navbar>")
        print("<h4>%s</h4>" % self.title)

        for name, ref in self.choices:
            print('<a href="%s">%s</a>' % (ref, name))

        print("</div>")


class Table:
    def __init__(self, width="100%", title=""):

        self.width = width
        self.cellspacing = 2
        self.cellpadding = 5
        self.title = title
        self.title_class = "title"
        self.table_class = "table"
        self.num_cols = 1

    def show(self):

        print("<table class=%s cellspacing=%d>" % (self.table_class, self.cellspacing))

        if self.title:
            print("<tr>")
            print("<th class=%s colspan=%d>" % (self.title_class, self.num_cols))
            print(self.title)
            print("</th>")
            print("</tr>")

        self.body()

        print("</table>")

    def body(self):
        pass

    def save(self, filename):
        tmp = sys.stdout
        sys.stdout = open(filename, "w")
        self.show()
        sys.stdout.close()
        sys.stdout = tmp


class ErrorMessage(Table):
    def __init__(self, msg="We have a problem", title="ERROR"):
        Table.__init__(self, title=title)

        self.msg = msg
        self.table_class = "errorTable"

    def body(self):
        print("<tr><td>%s</td></tr>" % self.msg)


class InfoMessage(Table):
    def __init__(self, msg="", title="INFORMATION", form=0):
        Table.__init__(self, title=title)

        self.form = form
        self.msg = msg
        self.table_class = "infoTable"

    def body(self):
        print("<tr><td>%s</td></tr>" % self.msg)


class Page:
    def __init__(self):

        sys.stderr = sys.stdout

        self.config = TransportConfig()
        self.sitedesc = self.config.get("DEFAULT", "sitedesc")

        self.refreshrate = 600
        self.stylesheet = "/stylesheet.css"
        self.title = self.sitedesc
        self.header = "DATA TRANSPORT NETWORK"
        self.doctype = '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">'

        self.navbar = Navbar()

    def headers(self):
        print("Content-type: text/html")

        if self.refreshrate:
            print("Refresh:", self.refreshrate)

    def head(self):
        print("<head>")
        print("<title>%s</title>" % self.title)
        print(
            '<style type="text/css"><!-- @import url(%s); --></style>' % self.stylesheet
        )
        print("</head>")

    def banner(self):

        print("<span id=title>%s</span><br>" % self.header)
        print("<span id=site>%s</span><br>" % self.sitedesc)

    def sidebar(self):

        if self.navbar:
            self.navbar.show()

    def body(self):
        pass

    def footer(self):
        return

        website = "http://transport.sri.com/TransportDevel"

        print('<a href="%s">Data Transport Network</a><br>' % website)
        print("<span>Copyright (C) 2002  SRI International</span>")

    def show(self):

        # print('Content-type: text/plain')
        # print()

        self.headers()

        print()
        print(self.doctype)
        print("<html>")
        self.head()
        print("<body>")

        print("<h1 id=banner>")
        self.banner()
        print("</h1>")

        print("<div id=content>")
        print("<div class=spacer></div>")

        print("<div id=sidebar>")
        self.sidebar()
        print("</div>")

        print("<div id=main>")
        print('<a name="pagetop"></a>')
        self.body()
        print("</div>")

        print("<div class=spacer></div>")
        print("</div>")

        print("<div id=footer>")
        self.footer()
        print("</div>")

        print("</body>")
        print("</html>")


def main():
    # Used as a component in other classes
    Page().show()



import os
import pathlib

from datatransport.utilities import remove_file

def test_remove_name_as_string(fs):
    filename = 'demo.txt'
    fs.create_file(filename)
    remove_file(filename)
    assert not os.path.exists(filename)
    
def test_remove_name_as_bytes(fs):
    filename = b'demo.txt'
    fs.create_file(filename)
    remove_file(filename)
    assert not os.path.exists(str(filename))
     
def test_remove_name_as_path(fs):
    filename = pathlib.Path('demo.txt')
    fs.create_file(filename)
    remove_file(filename)
    assert not os.path.exists(filename)

def test_remove_list(fs):
    filenames = ['a','b','c']
    for filename in filenames:
        fs.create_file(filename)
    remove_file(filenames)
    for filename in filenames:
        assert not os.path.exists(filename)

def test_remove_list_mixed(fs):
    filenames = ['a',b'b',pathlib.Path('c')]
    for filename in filenames:
        fs.create_file(filename)
    remove_file(filenames)
    for filename in filenames:
        assert not os.path.exists(str(filename))

def test_remove_missing(fs):
    filename = 'demo.txt'
    remove_file(filename)
    assert not os.path.exists(filename)



#!/usr/bin/env python

#######################################################
#
#   Simulated SBC data producer
#
#   Post data records into a news group.
#
#   2010-08-21  Todd Valentic
#               Initial implementation
#
#######################################################

from Transport      import ProcessClient
from Transport      import NewsPostMixin

import sys
import random

class SBC(ProcessClient,NewsPostMixin):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)
        NewsPostMixin.__init__(self)

        self.rate = self.getRate('rate')

    def run(self):

        while self.wait(self.rate):

            try:
                self.process()
            except:
                self.log.exception('Problem processing')

    def process(self):

        data = self.createData()

        self.newsPoster.postText(data,date=self.currentTime())

        self.log.debug('Posted data')

    def createData(self):

        fanState = random.randint(0,1)

        data = '[fan] %s' % fanState

        return data

if __name__ == '__main__':
    SBC(sys.argv).run()




#!/usr/bin/env python

from Transport      import ProcessClient
from Transport      import NewsPostMixin
from datetime       import datetime

import sys
import random
import ConfigParser
import StringIO
import pytz

class DataSource(ProcessClient,NewsPostMixin):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)
        NewsPostMixin.__init__(self)

        self.pollrate   = self.getDeltaTime('rate',60)
        self.sites      = self.get('sites','').split()
        self.config     = ConfigParser.ConfigParser()

        self.config.add_section('Temperature')
        self.config.add_section('Voltage')

    def run(self):

        internal    = 20
        external    = 40

        genseta     = 110
        gensetb     = 110

        while self.running:

            for site in self.sites:

                internal    += random.uniform(-2,2)
                external    += random.uniform(-2,2)
                genseta     += random.uniform(-1,1)
                gensetb     += random.uniform(-1,1)

                self.config.set('Temperature','Internal',internal)
                self.config.set('Temperature','External',external)
                self.config.set('Voltage','GeneratorA',genseta)
                self.config.set('Voltage','GeneratorB',gensetb)

                buffer = StringIO.StringIO()
                self.config.write(buffer)
                msg = buffer.getvalue()

                now = datetime.now(pytz.utc)
                headers = {'X-Site':site}

                self.newsPoster.postText(msg,date=now,headers=headers)

            self.log.info('Posted data samples')

            self.wait(self.pollrate)

if __name__ == '__main__':
    DataSource(sys.argv).run()




#!/usr/bin/env python3

##########################################################################
#
#   File post example
#
#   Periodically post a file to a news group 
#
#   2022-10-10  Todd Valentic
#               Initial implementation
#
##########################################################################

from datatransport import ProcessClient
from datatransport import NewsPostMixin 

import sys
import pathlib

class Client (ProcessClient, NewsPostMixin):

    def __init__(self, args):
        ProcessClient.__init__(self, args)
        NewsPostMixin.__init__(self)

        self.rate = self.get_rate('rate')
        self.text = self.get('text')

        self.log.info(f'Post message every {self.rate.period}')

    def run(self):

        filename = pathlib.Path('data.txt')

        while self.wait(self.rate):

            with filename.open(mode='w') as message:
                message.write(self.text)

            self.news_poster.post(str(filename))
            self.log.info('Posted message')

            filename.unlink()

if __name__ == '__main__':
    Client(sys.argv).run()




import pathlib

from datatransport.utilities import make_path 

def test_make_path_name_as_string(fs):
    filename = 'dir1/dir2/demo.txt'
    make_path(filename)
    assert pathlib.Path(filename).parents[0].exists() 
 
def test_make_path_name_as_path(fs):
    filename = pathlib.Path('dir1/dir2/demo.txt')
    make_path(filename)
    assert filename.parents[0].exists() 
  
def test_make_path_name_exists(fs):
    filename = pathlib.Path('dir1/dir2/demo.txt')
    fs.create_dir(filename.parents[0])
    make_path(filename)
    assert filename.parents[0].exists() 

def test_make_path_mode(fs):
    filename = pathlib.Path('dir1/dir2/demo.txt')
    make_path(filename, mode=0o777)
    assert filename.parents[0].stat().st_mode == 0o40777
    assert filename.parents[1].stat().st_mode == 0o40777
    
            



#!/usr/bin/env/python
"""ProcessClient"""

###########################################################################
#
#    ProcessClient
#
#    A base class for client programs of a process group.
#
#    History:
#
#    1.0.0  1999.??.??  TAV
#           Initial implementation.
#
#    1.2.0  2000.04.18  TAV
#           Migration of CORBA from Fnorb to omniORB.
#
#    1.2.1  2001-09-07  TAV
#           setStatus() -> setStatusCode()
#
#    1.2.2  2001-12-10  TAV
#           Added running data member here for general use.
#           Modified abort() to respect running and exit if stopped.
#
#    1.2.3  2001-12-26  TAV
#           Changed tav -> sri
#
#    1.2.4  2002-01-23  Todd Valentic
#           Added traceback() method to help in error reporting.
#           Added wait() method.
#           Added signal handling.
#
#    1.2.5  2002-08-28  Todd Valentic
#           sri.transport -> Transport
#
#    1.2.6  2002-12-19  Todd Valentic
#           Reworked the environment variable processing to mirror
#             that of the process group (.set and .add options).
#
#    1.2.7  2003-02-06  Todd Valentic
#           Added return value to wait (1=ok, 0=not running).
#
#    1.2.8  2003-06-11  Todd Valentic
#           Added createCallback method so other classes can use a
#             common system for plugins.
#
#    1.2.9  2003-06-20  Todd Valentic
#           Added default to createCallback.
#           Added createDeltaTime method.
#           Convert pollrate to int in wait - this lets us pass in
#            a DeltaDateTime object.
#
#    1.2.10 2003-06-27  Todd Valentic
#           Code cleanup. Renamed "datapath" to "workingdir".
#           Changed default values on abort.
#
#    1.2.11 2003-08-14  Todd Valentic
#           Changed createCallback and createDeltaTime to be mixin
#             classes and changed name to getCallback and
#             getDeltaTime to match consistency with other get*
#             methods.
#
#    1.2.12 2003-11-21  Todd Valentic
#           Fixed bug in wait() when sync enabled (pollrate needed to
#             converted to an int - see 1.2.9).
#
#    1.2.13 2003-12-23  Todd Valentic
#           Make sure to convert offset to int in wait() in case it
#             is a DateTime object.
#
#    1.2.14 2004-02-08  Todd Valentic
#           Removed unused imports (socket,CosNaming,DateTime)
#
#   2.0.0   2004-08-08  Todd Valentic
#           Converted to XML-RPC.
#           Major code cleanup.
#
#   2.0.1   2004-12-27  Todd Valentic
#           Convert from mx.DateTime to datetime
#           Added client.name to config options
#
#   2.0.2   2005-04-04  Todd Valentic
#           Added multiple conf files.
#
#   2.0.3   2005-04-10  Todd Valentic
#           Added is_running() and is_stopped() methods
#
#   2.0.4   2005-04-17  Todd Valentic
#           Added put() method for adding to config space.
#
#   2.0.5   2005-06-22  Todd Valentic
#           Changed default PATH to be /usr/local/bin:/bin:/usr/bin
#               This better matches the default shell version.
#
#   2.0.6   2005-11-12  Todd Valentic
#           Pass through keywords in get*() methods.
#
#   2.0.7   2005-11-30  Todd Valentic
#           Show configuration file names in debug log output.
#
#   2.0.8   2006-01-18  Todd Valentic
#           Added exit() command to compliment abort().
#
#   2.0.9   2006-02-03  Todd Valentic
#           Start each client in group.work instead of group.home.
#               This move separates the config files from the local
#               runtime generated files. Makes distributing groups
#               much easier.
#
#   2.0.10  2006-10-25  Todd Valentic
#           Added sleep parameter to wait(). This lets alternative
#               methods be used besides time.sleep(). In particular
#               this is used by threads who need Event.wait().
#
#   2.0.11  2006-10-26  Todd Valentic
#           Added currentTime() and utc() methods.
#
#   2.0.12  2007-03-17  Todd Valentic
#           Use os.path.sep
#           Make sure to read config files in path.group
#
#   2.0.13  2009-05-03  Todd Valentic
#           Added path.exec config option
#
#   2.0.14  2010-05-25  Todd Valentic
#           Because stop() now blocks, we cannot issue logout() in the
#               signal handler. Not really needed anymore.
#
#   2.0.15  2010-08-16  Todd Valentic
#           Add support to wait() for the ConfigMixin.Rate object.
#
#   2016-07-10  Todd Valentic
#               Use atexit to call logoutclient()
#
#   2016-12-23  Todd Valentic
#               Use datetime.timezone.utc for UTC
#
#   2020-01-08  Todd Valentic
#               Add atStart support to wait()
#
#   2020-10-09  Todd Valentic
#               Python3: xmlrpc.client
#               Drop datefunc (replace with new standard methods)
#
#   2021-07-13  Todd Valentic
#               Add show_thread option for logging
#
#   2022-10-07  Todd Valentic
#               Fix wait()
#                   Signal handling changed in Python 3.5 to that
#                   sleep.time() isn't interrupted. We had relied on
#                   the interrupt to exit on demand. Rework to use
#                   threading.Event(). For details, see:
#
#                   "break-interrupt-a-time-sleep-in-python"
#                       https://stackoverflow.com/a/46346184
#
#                   https://peps.python.org/pep-0475/
#
#               Make running a property
#               Convert API to snake case naming
#                   CurrentTime -> current_time
#
#   2023-07-09  Todd Valentic
#               Use TransportConfig find_config_files()
#
#   2023-07-25  Todd Valentic
#               Add set() and options() to config proxy
#               Do not map getters from config, use config object               
#               current_time() renamed to now()
#
#   2026-03-05  Todd Valentic
#               Expand user in PATH components
#
###########################################################################

import atexit
import datetime
import fnmatch
import glob
import logging
import os
import signal
import sys
import threading
import time
import xmlrpc.client

from functools import partial
from logging.handlers import RotatingFileHandler
from logging.handlers import SocketHandler

import sapphire_config as sapphire

from . import Root
from . import TransportConfig

# For the get methods forwarded to the config
# pylint: disable=no-member


class ProcessClient(Root):
    """ProcessClient"""

    def __init__(self, argv):
        self.groupname = argv[1]
        self.name = argv[2]
        self.utc = datetime.timezone.utc
        self.subprocs = set()
        self.started_rates = {}
        self.exit_event = threading.Event()
        config_files = self.load_config()

        url = self.config.parser.get("TransportServer", "url")
        self.server = xmlrpc.client.ServerProxy(url)

        self.setup_signals()
        self.setup_log()
        self.setup_environment()
        self.setup_working_dir()

        self.server.loginclient(self.groupname, self.name, os.getpid())

        self.log.debug("Configuration files:")
        for filename in config_files:
            self.log.debug("  - %s", filename)

        atexit.register(self.on_exit)

    @property
    def running(self):
        """Indictate if process should continue running"""
        return not self.exit_event.is_set()

    def on_exit(self):
        """Called when process exits"""
        if not self.is_running():
            self.server.logoutclient(self.groupname, self.name, os.getpid())

    def setup_signals(self):
        """Setup signal handlers"""

        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGHUP, self.signal_handler)

        # Restore the normal child signal handling. The TransportServer
        # overrides this signal to ignore child process return codes
        # preventing zombie (defunct) processes. However, this also
        # has the effect of preventing the commands.* methods from
        # working as expected. Restoring the default behaviour returns
        # things to normal.

        signal.signal(signal.SIGCHLD, signal.SIG_DFL)

    def signal_handler(self, signum, _frame):
        """Handle signal even, exit if requested"""

        if self.is_running():
            self.exit_event.set()
            self.log.info("Exit request")

            for proc in self.subprocs:
                if proc.poll() is None:
                    proc.send_signal(signum)

    def _setup_log_socket_handler(self, formatter):
        """Add a socket log handler"""

        host = self.config.get("log.socket.host", fallback="localhost")
        port = self.config.get_int("log.socket.port", fallback=9020)

        socket_handler = SocketHandler(host, port)
        socket_handler.setFormatter(formatter)

        return socket_handler

    def _setup_log_file_handler(self, formatter):
        """Add a rotating file log handler"""

        filename = self.config.get("log.file")
        maxbytes = self.config.get_bytes("log.maxbytes", fallback="100kb")
        backup_count = self.config.get_int("log.backupcount", fallback=3)

        rotating_handler = RotatingFileHandler(filename, "a", maxbytes, backup_count)
        rotating_handler.setFormatter(formatter)

        return rotating_handler

    def setup_log_formatter(self):
        """Initialize a log formatter"""

        msgfmt = "[%(asctime)s.%(msecs)03d %(levelname)7s] %(name)s"

        if self.config.get_boolean("log.showthread", fallback=False):
            msgfmt += " [%(threadName)s]"

        msgfmt = msgfmt + ": %(message)s"

        datefmt = "%Y-%m-%d %H:%M:%S"

        return logging.Formatter(msgfmt, datefmt)

    def setup_log(self):
        """Setup log handlers"""

        level = self.config.get("log.level", "info")
        formatter = self.setup_log_formatter()

        root_logger = logging.getLogger("")
        self.log = logging.getLogger(f"{self.groupname}/{self.name}")

        if self.config.get_boolean("log.file.enable", True):
            root_logger.addHandler(self._setup_log_file_handler(formatter))

        if self.config.get_boolean("log.socket.enable", True):
            root_logger.addHandler(self._setup_log_socket_handler(formatter))

        if level == "warning":
            root_logger.setLevel(logging.WARNING)
        elif level == "info":
            root_logger.setLevel(logging.INFO)
        else:
            root_logger.setLevel(logging.DEBUG)

    def load_config(self):
        """Load the configuration files"""

        defaults = {
            "client.name": self.name,
            "group.name": self.groupname,
            "group.basename": os.path.basename(self.groupname),
            "group.dirname": os.path.dirname(self.groupname),
        }

        config = TransportConfig(defaults)
        basepath = config["TransportServer"].get_path("path.groups")
        hostname = config["TransportServer"].get("hostname")
        config_files = config.find_config_files(self.groupname, basepath, hostname)

        # Read one at a time so we know if one has an error

        for filename in config_files:
            config.read(filename)

        self.config = config[self.name]
        self.config.set = self.config.__setitem__
        self.config.options = self.config._options
        self.config.get_components = partial(self.config.get_components, parent=self)

        self.hostname = config.get("DEFAULT", "hostname")

        return config_files

    def setup_environment(self):
        """Setup up environment variables"""

        self.log.debug("Setting environment variables:")

        paths = self.config.get("path.exec", "/usr/local/bin:/bin:/usr/bin")
        paths += ":" + self.config.get("group.bin")
        paths += ":" + self.config.get("path.bin")

        os.environ["PATH"] = paths 

        options = self.config.options()

        setoptions = fnmatch.filter(options, "environ.set.*")
        addoptions = fnmatch.filter(options, "environ.add.*")

        for option in setoptions:
            value = self.config.get(option)
            var = option.split(".")[2].upper()
            os.environ[var] = value
            self.log.debug("  set: [%s] = %s", var, value)

        for option in addoptions:
            value = self.config.get(option)
            var = option.split(".")[2].upper()
            if var in os.environ:
                # only add if not there
                if os.environ[var].find(value) == -1:
                    os.environ[var] += ":" + value
            else:
                os.environ[var] = value
            self.log.debug("  add: [%s] %s", var, value)

        paths = os.environ["PATH"].split(":")
        expanded_path = [os.path.expanduser(p) for p in paths]
        os.environ["PATH"] = ":".join(expanded_path)

        keys = sorted(os.environ.keys())

        self.log.debug("Current environment:")

        for key in keys:
            self.log.debug("  [%s] = %s", key, os.environ[key])

    def setup_working_dir(self):
        """Change to working directory"""

        workingdir = os.path.join(self.config.get("group.work"), self.name)

        if not os.path.exists(workingdir):
            os.makedirs(workingdir)

        os.chdir(workingdir)

    def abort(self, msg="Exiting", status=1):
        """Exit process with error"""
        self.log.error(msg)
        self.exit_event.set()
        sys.exit(status)

    def exit(self, msg="Exiting", status=0):
        """Exit process"""
        self.log.info(msg)
        self.exit_event.set()
        sys.exit(status)

    def stop(self):
        """Request to stop"""
        self.exit_event.set()

    def is_running(self):
        """Indicate the process is running"""
        return self.running

    def is_stopped(self):
        """Indicate the processes has stopped"""
        return not self.running

    def now(self):
        """Return current time as datetime with UTC timezone"""
        return datetime.datetime.now(self.utc)

    def wait(self, pollrate, offset=None, sync=False):
        """Wait for a given time, short circuit if we have stopped running"""

        if not self.running:
            return 0

        if isinstance(pollrate, sapphire.Rate):
            period = pollrate.period
            sync = pollrate.sync
            offset = pollrate.offset

            # Short circuit first time if at_start is set

            if pollrate.at_start:
                pollrate.at_start = False
                return self.running
        else:
            period = pollrate

        if isinstance(period, datetime.timedelta):
            secs = period.total_seconds()
        else:
            secs = int(period)

        if offset is None:
            offset_secs = 0
        elif isinstance(offset, datetime.timedelta):
            offset_secs = offset.total_seconds()
        else:
            offset_secs = float(offset)

        if sync:
            curtime = time.time() - offset_secs
            waittime = max(0, secs - curtime % secs + 0.000)
        else:
            waittime = secs

        self.exit_event.wait(waittime)

        return self.is_running()

    def main(self):
        """Main application"""

    def run(self):
        """Run main application, capture any errors to log"""
        try:
            self.init()
        except Exception:  # pylint: disable=broad-exception-caught
            self.log.exception("Problem detected in init")
            self.abort()

        try:
            self.main()
        except Exception:  # pylint: disable=broad-exception-caught
            self.log.exception("Problem detected in main")
            return  # No clean exit so the watchdog will restart us

        self.exit()



#!/usr/bin/env python
"""Pattern Template"""

###########################################################################
#
#   Pattern Template
#
#   This class provides a small template system for replacing content
#   in strings.
#
#   2001-10-11  Todd Valentic
#               Initial implementation.
#
#   2004-04-10  Todd Valentic
#               Use string methods vs string module
#               Use isinstance vs type module
#
#   2005-06-15  Todd Valentic
#               Added ability for value to be a dictionary.
#               Added setValue for caching of value parameter.
#
#   2012-02-10  Todd Valentic
#               Cache the value if given in the call.
#
#   2022-10-12  Todd Valentic
#               Python3 port
#               Use parse_slice (https://stackoverflow.com/a/68899290)
#               Rework implementation with regex instead of eval
#               setValue -> set_value
#
#   2023-07-26  Todd Valentic
#               Missing check for pattern 
#
#   SPDX-License-Identifier: GPL-3.0-or-later
#   Copyright (C) 1999-2022 Todd Valentic
#
###########################################################################

import re

from typing import Union

re_pattern = re.compile(r'(?<=<).*?(?=>)')
re_slice = re.compile(r'^.*\[(-?[\d]*):(-?[\d]*)[:]?(-?[\d]*)\]$')
re_index = re.compile(r'^.*\[(-?[\d]+)\]$')
re_key = re.compile(r'^.*\[\"(.+)\"\]$')

def parse_slice(string: str) -> slice:
    """
    Parse a string representation of a slice and return a slice object
    """
    # Matches one required colon, one optional colon, and up to three
    # positive or negative numbers between them (i.e. name[1:3])
    match = re_slice.match(string)
    if match:
        args = [int(s) if s else None for s in  match.group(1, 2, 3)]
        return slice(*args)
    raise ValueError("Could not parse slice")

def parse_index(string: str) -> int:
    """
    Parse a string representation of a list index and return an int
    """
    match = re_index.match(string)
    if match:
        return int(match.group(1))
    raise ValueError('Count not parse index')

def parse_key(string: str) -> str:
    """
    Parse a string representation of a dict key (i.e., name[key])
    """
    match = re_key.match(string)
    if match:
        return match.group(1)
    raise ValueError('Count not parse key')

def parse_patterns(string: str) -> list:
    """
    Return a unique list of patterns (<name>) found in the string
    """
    return list(set(re_pattern.findall(string)))

class PatternTemplate:
    """String pattern replacement"""

    def __init__(self, pattern: str, sep: str=None):
        self.pattern = pattern
        self.sep = sep
        self.value = None

    def set_value(self, value: Union[str, list, dict]) -> None:
        """Cache the value of the replacement string"""

        self.value = value

    def lookup_value(self, entry: str, value: Union[str, list, dict], sep: str='') -> str:
        """Look up the entry in the value str, list or dict"""

        if isinstance(value, (list, tuple)):

            try:
                idx = parse_index(entry)
                return value[idx]
            except ValueError:
                pass

            try:
                s = parse_slice(entry)
                return sep.join(value[s])
            except ValueError:
                pass

            return sep.join(value)

        if isinstance(value, dict):

            try:
                key = parse_key(entry)
                return value[key]
            except ValueError:
                pass

            return entry

        return value


    def __call__(self, src: str, value: str=None) -> str:
        """Replace <pattern> with value"""

        if value is not None:
            self.set_value(value)

        value = self.value

        if value is None:
            return src

        if self.sep:
            value = value.split(self.sep)

        for entry in parse_patterns(src):
            if not entry.startswith(self.pattern):
                continue
            v = self.lookup_value(entry, value, self.sep)
            src = src.replace(f'<{entry}>', str(v))

        return src



#!/usr/bin/env python

from Transport  import ProcessClient
from Transport  import NewsPostMixin
from datetime   import datetime

import sys

class Source (ProcessClient,NewsPostMixin):
    
    def __init__(self,argv):
        ProcessClient.__init__(self,argv)
        NewsPostMixin.__init__(self)

        self.rate = self.getDeltaTime('post.rate',60)

        self.log.info('rate: %s' % self.rate)

    def run(self):

        while self.wait(self.rate,sync=True):   
            self.newsPoster.postText('[%s] %s' % (self.name,datetime.now()))
            self.log.info('Posted message')

if __name__ == '__main__':
    Source(sys.argv).run()




#!/usr/bin/env python3
"""Cancel messages in a new group"""

##########################################################################
#
#   Cancel messages in a news group
#
#   2022-10-10  Todd Valentic
#               Add header comment block
#               Make a package command
#
##########################################################################

import argparse

from datatransport import newstool


def main():
    """Script entry point"""

    parser = argparse.ArgumentParser()

    parser.add_argument("newsgroup")

    parser.add_argument(
        "-s",
        "--server",
        default="localhost",
        help="News server host (default %(default)s)",
    )

    parser.add_argument(
        "-p",
        "--port",
        default=119,
        type=int,
        help="News server port (default %(default)s)",
    )

    args = parser.parse_args()

    server = newstool.NewsControl()
    server.set_server(args.server, port=args.port)

    if not server.has_newsgroup(args.newsgroup):
        print(f"Error: newsgroup does not exist: {args.newsgroup}")
        return 1

    print(f"Canceling messages in {args.newsgroup}")

    server.cancel_newsgroup(args.newsgroup)

    return 0



#!/usr/bin/env python

##########################################################
#
#   Hello World
#
#   Print a message to the log
#
#   2010-08-16  Todd Valentic
#               Initial implementation
#
##########################################################

from Transport import ProcessClient

import sys

class HelloWorld (ProcessClient):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)

        self.rate    = self.getRate('rate','1:00')
        self.message = self.get('message')

        if self.message is None:
            self.abort('You need to specify a message')

        self.log.info('Rate: %s' % self.rate)

    def run(self):

        while self.wait(self.rate):
            self.log.info(self.message)

if __name__ == '__main__':
    HelloWorld(sys.argv).run()




#!/usr/bin/env python3
"""Text polling example"""

##########################################################################
#
#   Polling example
#
#   Periodically poll for messages from a news group
#
#   2022-10-10  Todd Valentic
#               Initial implementation
#
##########################################################################

import sys

from datatransport import ProcessClient
from datatransport import NewsPoller


class Client(ProcessClient):
    """Process client"""

    def __init__(self, args):
        ProcessClient.__init__(self, args)

        self.poller = NewsPoller(self, callback=self.process)
        self.main = self.poller.main

        self.log.info("Poll for messages")

    def process(self, message):
        """Process handler"""

        text = message.get_payload()
        self.log.info('Received: "%s"', text)


if __name__ == "__main__":
    Client(sys.argv).run()



#!/usr/bin/env python
"""NewsPoller"""

###########################################################################
#
#   Data Transport NewsPoller
#
#   Poll a newsgroup and call a callback function when new messages arrive.
#
#   History:
#
#   1999-??-??  Todd Valentic
#               Initial implementation.
#
#   2000-04-29  Todd Valentic
#               Added the groupExists() check at the top of the run() method.
#               If the group isn't available (either it doesn't exist or
#               the server is down) then stall and retry every minute. This
#               situation often ocurrs when a new process group is added to
#               the system and the newsgroups don't exist yet. The posting
#               processes will create the groups, but they may not show up
#               for a few minutes. By stalling and rechecking, we eventually
#               sync up.
#
#   2001-08-11  Todd Valentic
#               Added the 'text' parameter to handle ProcessText callbacks.
#
#   2001-08-13  Todd Valentic
#               Added the idle function to get called during the main loop.
#
#   2001-11-13  Todd Valentic
#               Added self.running check to run().
#
#   2001-12-10  Todd Valentic
#               Added running check before sleeping.
#               ProcessClient now defines self.running.
#
#   2001-12-26  Todd Valentic
#               Changed tav -> sri
#
#   2002-02-24  Todd Valentic
#               Major code cleanup.
#               Changed the default config names be prefixed with "poll.". This
#                   prefix is also user setable in __init__.
#               Default server is now localhost.
#
#   2002-03-05  Todd Valentic
#               Added exit_on_error parameter (replaces debug).
#
#   2002-04-23  Todd Valentic
#               Changed call from newsPoller.prifix -> news_poll_prefix.
#                   The former is only used to prefix filenames.
#
#   2002-08-28  Todd Valentic
#               sri.transport -> Transport
#
#   2003-04-15  Todd Valentic
#               Added support for NewsPoller to listen to isRunning.
#               Removed the local system() call. This had added the
#                   current working directory to the path. This is
#                   now done in the configuration files.
#
#   2003-06-30  Todd Valentic
#               Change to new single callback method for processing.
#
#   2003-08-11  Todd Valentic
#               Removed unneeded 'import os'
#               Made rate a DeltaTime parameter.
#               Modified to handle multiple input newsgroups. newsPoller is
#                   now a list instead of a single instance.
#
#   2003-08-14  Todd Valentic
#               createDeltaTime -> getDeltaTime
#               createCallback -> getCallback
#
#   2004-02-05  Todd Valentic
#               Changed idleFunc -> idle. Keep older version for awhile.
#
#   2004-08-17  Todd Valentic
#               Changed to new logger interface.
#
#   2005-01-06  Todd Valentic
#               Changed NewsTool import.
#
#   2005-04-10  Todd Valentic
#               Use is_stopped() as poller stop function.
#
#   2005-05-09  Todd Valentic
#               Generalized run() method by adding _run_pollers and _run_idle.
#                   This makes is easier for derived classes (ie SyncPoller)
#                   to override these functions.
#
#   2005-07-27  Todd Valentic
#               New ability to specify multiple newsgroups with wild
#                   cards in their names. This requires a connection
#                   to the newsserver be available. If no
#
#   2005-11-30  Todd Valentic
#               Added debug listing of news pollers at creation.
#
#   2006-05-31  Todd Valentic
#               Added ability to set port for news server in config file.
#
#   2007-02-05  Todd Valentic
#               Added catchup.reset - when True, the catchup is applied each
#                   polling cycle. You can use this, for example, to only
#                   processes the last message.
#
#   2007-04-15  Todd Valentic
#               Moved catchup.reset processing into _run_pollers(). We
#                   were crashing on NNTP temperorary errors.
#
#   2008-08-11  Todd Valentic
#               Make sure we can connect to the news server at start.
#
#   2008-08-12  Todd Valentic
#               Correctly implement feature in 2.2.13.
#
#   2008-08-20  Todd Valentic
#               Make poll.rate.sync a boolean.
#               Make poll.rate.offset a timedelta.
#               Fix catchup.reset processing (broken in 2.2.12)
#
#   2014-01-17  Todd Valentic
#               Add poll.newsgroups.exclude
#
#   2020-01-08  Todd Valentic
#               Use rate object for poll config
#               Set lastReadPrexix if used in ConfigComponent object
#
#   2022-10-19  Todd Valentic
#               Remove deprecated idleFunc
#               Port to transport3 / python3
#
#   2023-07-04  Todd Valentic
#               Convert from being a mixin class
#
#   2023-07-27  Todd Valentic
#               Use AccessMixin
#
###########################################################################

from datatransport import newstool
from datatransport import AccessMixin


class NewsPoller(AccessMixin):
    """News poller"""

    def __init__(self, parent, prefix="poll", callback=None, idle=None):
        AccessMixin.__init__(self, parent)

        self.prefix = prefix
        self.idle = idle
        self.rate = self.config.get_rate(f"{prefix}.rate", 60)

        self.news_pollers = self.create_pollers(prefix, callback)

    # pylint: disable=inconsistent-return-statements
    def connect_to_server(self, host, port):
        """Open a connection to the news server"""

        server = newstool.NewsTool()
        server.set_server(host, port)

        while self.is_running():
            try:
                server.open_server()
                return server
            except:  # pylint: disable=bare-except
                self.log.exception(f"Failed to connect to {host}:{port}")
                self.wait(15)

        self.abort("Exiting")

    def get_newsgoups(self, prefix, server):
        """Get a list of newsgroups"""

        targetgroup = self.config.get(f"{prefix}.newsgroup", "")
        targetgroups = self.config.get_list(f"{prefix}.newsgroups", targetgroup)
        excludegroup = self.config.get(f"{prefix}.newsgroup.exclude", "")
        excludegroups = self.config.get_list(f"{prefix}.newsgroups.exclude", excludegroup)

        newsgroups = []

        for groupspec in targetgroups:
            if "*" in groupspec or "?" in groupspec:
                try:
                    matches = server.list_newsgroups(groupspec, excludegroups)
                    newsgroups.extend(matches)
                except:  # pylint: disable=bare-except
                    self.log.info(f"Problem matching '{groupspec}'")
                    continue
            else:
                newsgroups.append(groupspec)

        return newsgroups

    def create_pollers(self, prefix, callback):
        """Create a news poller for each newsgroup"""

        host = self.config.get(f"{prefix}.newsserver", "localhost")
        port = self.config.get_int(f"{prefix}.newsserver.port", 119)

        exit_on_error = self.config.get_boolean(f"{prefix}.exit_on_error", False)
        retry_wait = self.config.get_timedelta(f"{prefix}.retry_wait", 60)

        # Prefix the tracking file when used in ConfigComponent
        # because we might have multiple references to the same group

        if hasattr(self.parent, "prefix") and hasattr(self.parent, "name"):
            last_read_prefix = f"{self.parent.prefix}-{self.parent.name}"
        else:
            last_read_prefix = ""

        server = self.connect_to_server(host, port)
        newsgroups = self.get_newsgoups(prefix, server)

        self.log.debug("Creating news pollers:")

        pollers = []

        for newsgroup in newsgroups:
            poller = newstool.NewsPoller()
            poller.set_server(host, port)
            poller.set_newsgroup(newsgroup)
            poller.set_log(self.log)
            poller.set_callback(callback)
            poller.set_stop_func(self.is_stopped)
            poller.set_retry_wait(retry_wait.total_seconds())
            poller.set_debug(exit_on_error)
            poller.set_last_read_prefix(last_read_prefix)

            pollers.append(poller)

            self.log.debug(" - [%s:%s] %s", host, port, newsgroup)

        return pollers

    def run_pollers(self, catchup, reset):
        """Run each poller"""

        for poller in self.news_pollers:
            if reset and catchup != 0:
                poller.mark_read(catchup, reset=reset)

            poller.poll()

    def run_idle(self):
        """Run the idle handler"""

        if self.idle and self.is_running():
            self.idle()

    def mark_read(self, catchup, reset):
        """Mark each newsgroup as read"""

        for poller in self.news_pollers:
            while not poller.has_newsgroup():
                self.self.log.error(
                    f"The polling group ({poller.newsgroup_header}) "
                    f"does not exist on "
                    f"{poller.server_host}:{poller.server_port}, "
                    f"waiting 60 seconds ..."
                )
                if not self.wait(60):
                    return

            if catchup != 0:
                poller.mark_read(catchup, reset)

    def process(self):
        """Process messages"""

        catchup = self.config.get_int(f"{self.prefix}.catchup", 0)
        catchup_reset = self.config.get_boolean(f"{self.prefix}.catchup.reset", False)
        exit_on_error = self.config.get_boolean(f"{self.prefix}.exit_on_error", False)

        if catchup == 0:
            self.log.debug("Starting with first article in last read file")
        else:
            self.mark_read(catchup, catchup_reset)

        while self.is_running():
            try:
                self.run_pollers(catchup, catchup_reset)
            except:  # pylint: disable=bare-except
                self.log.exception("Error detected during polling")
                if exit_on_error:
                    self.abort("Exiting on error")

            try:
                self.run_idle()
            except:  # pylint: disable=bare-except
                self.log.exception("Error detected during idle")
                if exit_on_error:
                    self.abort("Exiting on error")

            yield 1

    def main(self):
        """Main loop"""

        if not self.wait(self.rate):
            return

        for _step in self.process():
            self.wait(self.rate)

    def run_step(self):
        """Run one step in main loop"""

        next(self.process())



#!/usr/bin/env python

##############################################################
#
#   ConfigComponents Demonstration
#
#   1.0.0   2004-02-01  Todd Valentic
#           Initial implementation.
#
##############################################################

from Transport import ProcessClient
from Transport import ConfigComponent
from Transport import NewsPostMixin

import sys

class Watch(ConfigComponent,NewsPostMixin):

    def __init__(self,name,parent):
        ConfigComponent.__init__(self,'watch',name,parent)
        NewsPostMixin.__init__(self)

        self.path    = self.get('path','.')
        self.label   = self.get('label','')
        self.host    = self.get('host','')
        self.option  = self.get('option','')
        self.desc    = self.get('desc','')
        self.subject = self.get('subject','')

        self.log.info('Component: %s' % name)
        self.log.info('  path:    %s' % self.path)
        self.log.info('  label:   %s' % self.label)
        self.log.info('  host:    %s' % self.host)
        self.log.info('  option:  %s' % self.option)
        self.log.info('  desc:    %s' % self.desc)
        self.log.info('  subject: %s' % self.subject)

class Demo(ProcessClient):

    def __init__(self,args):
        ProcessClient.__init__(self,args)

        self.components = self.getComponents('watches',Watch)

    def run(self):
        self.exit()

if __name__ == '__main__':
    Demo(sys.argv).run()




#!/usr/bin/env python
"""ConfigComponent"""

##########################################################################
#
#   ConfigComponent
#
#   DataTransport version of the Sapphire ConfigComponent.
#
#   2023-06-28  Todd Valentic
#               Initial implementation. Based on transport v2 version.
#
##########################################################################

from functools import partial
from .accessmixin import AccessMixin

import sapphire_config as sapphire


class ComponentLog:
    """Prefix log entries with the component name"""

    def __init__(self, name, log):
        self.name = name
        self.log = log

    def __getattr__(self, name):
        def inner(msg, *args):
            fn = getattr(self.log, name)
            fn(f"[{self.name}] {msg}", *args)

        return inner


class ConfigComponent(sapphire.Component, AccessMixin):
    """Component class for config file entries"""

    # pylint: disable=too-many-arguments

    def __init__(self, prefix, name, config, parent, **kw):
        sapphire.Component.__init__(self, prefix, name, config, parent, **kw)
        AccessMixin.__init__(self, parent)

        self.log = ComponentLog(f"{prefix}.{name}", parent.log)

        self.config.get_components = partial(self.config.get_components, parent=self)



#!/usr/bin/env python3
"""Request generator"""

##########################################################################
#
#   Generate group control request messages
#
#   2023-07-27  Todd Valentic
#               Initial implementation
#
##########################################################################

import sys

from datatransport import ProcessClient
from datatransport import NewsPoster

class RequestGenerator(ProcessClient):
    """Process client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.poster = NewsPoster(self)
        self.rate = self.config.get_rate('rate', 60)

    def main(self):
        
        states = ['start', 'stop']
        state = 0

        while self.wait(self.rate):
            msg = f"{states[state]}=hello"
            self.poster.post_text(msg)
            self.log.info('Posted %s', msg) 
            state = (state+1) % len(states)

if __name__ == '__main__':
    RequestGenerator(sys.argv).run()




#!/usr/bin/env python
"""Test watchdog restart"""

############################################################################
#
#   Crasher
#
#   This script crashs in __init__() to test the transport server's ability 
#   to capture and report stderr to the server's log.
#
#   History:
#
#   2023-07-06  Todd Valentic
#               Initial implementation
#
############################################################################

import sys

from datatransport import ProcessClient


class InitCrasher(ProcessClient):
    """Process that unexpectedly dies"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        raise IOError("Goes boom in __init__()")

if __name__ == "__main__":
    InitCrasher(sys.argv).run()



#!/usr/bin/env python
"""datetime helper functions"""

##########################################################################
#
#   datetime helper functions.
#
#   Most of these functions are now available as built-in methods since
#   it was originally written. It remains here to allow for existing
#   code that uses it to run unchanged.
#
#   2004-12-27  Todd Valentic
#               Initial implementation
#
#   2005-04-07  Todd Valentic
#               Added parse_timedelta
#
#   2020-10-08  Todd Valentic
#               Updated for python3
#               Replaced a number of functions with built-ins
#               Use pytimeparse2
#
#   SPDX-License-Identifier: GPL-3.0-or-later
#   Copyright (C) 1999-2022 Todd Valentic
#
##########################################################################

from datetime import datetime, timedelta, timezone

import pytimeparse2 as pytimeparse

def datetime_as_seconds(dt: datetime):
    """Return POSIX timestamp"""

    return dt.timestamp()

def strptime(datestr: str, fmt: str, tzinfo: timezone=None):
    """datetime parsed from a string"""

    return datetime.strptime(datestr, fmt).replace(tzinfo=tzinfo)

def timedelta_as_seconds(td: timedelta):
    """Return timedelta as seconds"""

    return td.total_seconds()

def parse_timedelta(timestr: str):
    """timedelta parsed from a string"""

    return timedelta(seconds=pytimeparse.parse(timestr))



#!/usr/bin/env python3
"""Transport command line interface"""

############################################################################
#
#   transportctl
#
#   This script controls the transport server from the command line.
#
#   History:
#
#   1.0.0   1999.??.??  TAV
#           Initial implementation
#
#   1.2.0   2000.04.20  TAV
#           Updated to use omniORB for CORBA functions.
#
#   1.3.0   2000-06-22  TAV
#           Added ability to change options for the web-site users.
#
#   1.3.1   2000-08-08  TAV
#           Fixed bug if no option given on command line, call help.
#
#   1.3.2   2001-08-29  TAV
#           Handle new list of config files in the ProcessGroupInfo struct.
#
#   1.3.3   2001-10-23  TAV
#           Fixed a bug in the "user set" command. The user's values were
#               being reset to the defaults with each change.
#
#   1.3.4   2002-02-11  Todd Valentic
#           Added ability to pass args to start client command.
#
#   1.3.5   2002-08-28  Todd Valentic
#           Setup via configure script
#           sri.transport -> Transport
#           Removed web-site user functions.
#
#   1.3.6   2004-08-16  Todd Valentic
#           Cleanup for XML-RPC.
#           Removed log rotation.
#           Removed TransportClient.
#
#   2016-12-23  Todd Valentic
#               Migrate to datatransport package format
#
#   2020-10-09  Todd Valentic
#               Python3: xmlrpc.client
#               Python3: print
#
#   2023-08-17  Todd Valentic
#               Improve error checking
#               PEP8 compliance
#
#   2023-08-29  Todd Valentic
#               Missing f-string in list clients
#
############################################################################

import json
import os
import shutil
import subprocess
import sys
import time
import xmlrpc.client

from pathlib import Path

from datatransport import TransportConfig


class TransportControl:
    """Transport Control client"""

    def __init__(self):
        self.command = {
            "help": self.help,
            "list": self.list,
            "status": self.show_status,
            "cleanup": self.cleanup,
            "reload": self.reload,
            "add": self.add_group,
            "remove": self.remove_group,
            "restart": self.restart,
            "start": self.start,
            "stop": self.stop,
            "server": self.server_control,
        }

    def help(self):
        """Show help message"""

        print()
        print("-" * 70)
        print("Transport Control Commands:")
        print("-" * 70)
        print("\thelp                     - Show this page")
        print("\tstatus                   - Transport server status")
        print("\tloglevel                 - Set log level on group or server")
        print("\treload <group>           - Reload group's config file")
        print("\treload server            - Reload process groups")
        print("\tlist                     - List all process groups")
        print("\tlist <group>             - List clients in a group")
        print("\tstart <group>            - Start a process group")
        print("\tstop <group>             - Stop a process group")
        print("\trestart <group>          - Stop/start a process group")
        print("\tadd <group>              - Notify server of new group")
        print("\tremove <group>           - Remove a group from the server")
        print("\tserver start             - Start the transport server")
        print("\tserver stop              - Stop the transport server")
        print("\tstart <group> <client>   - Start the client in group")
        print("\tstop <group> <client>    - Stop the client in group")
        print("\tcleanup                  - Remove processes after crash")
        print("-" * 70)
        print()

    def run_command(self, argv):
        """Run command"""

        if len(argv) < 2:
            self.help()
            sys.exit(1)

        cmd = argv[1]
        arg = argv[2:]

        if cmd not in self.command:
            print("Unknown command:", cmd)
            self.help()
            return 1

        if cmd == "cleanup":
            self.cleanup()
            return 0

        if cmd == "help":
            self.help()
            return 0

        url = TransportConfig().get("TransportServer", "url")
        self.server = xmlrpc.client.ServerProxy(url)

        if cmd != "server":
            # Skip if we need to start the server

            try:
                self.status = self.server.status()
            except (ConnectionRefusedError, xmlrpc.client.Error) as err:
                print(f"Cannot connect to the transport server: {err}")
                return 1

        return self.command[cmd](arg)

    def has_group_or_exit(self, group):
        """Exit if group does not exist"""

        try:
            groups = self.server.listgroups()
        except xmlrpc.client.Error as err:
            print(f"Problem getting group list: {err}")
            sys.exit(1)

        if group in groups:
            return

        print(f'The process group "{group}" does not exist')
        print('  Use the "list" command to print the available groups')

        sys.exit(1)

    def has_client_or_exit(self, group, client):
        """Exit if group or client do not exist"""

        self.has_group_or_exit(group)

        clients = self.server.listclients(group)

        if client in clients:
            return

        print(f'The client "{client}" does not exist in {group}')
        print('  Use the "list" command to print the available clients')

        sys.exit(1)

    def show_status(self, _arg):
        """Show server status"""

        print("Server status:")
        print(json.dumps(self.status, sort_keys=True, indent=4))

        return 0

    def reload(self, arg):
        """Reload a process group or server"""

        if not arg:
            print('You need to specify a process group or "server"')
            return 1

        if arg[0] == "server":
            self.server.reloadgroups()
            return 0

        group = arg[0]

        self.has_group_or_exit(group)

        try:
            self.server.reloadgroup(group)
        except xmlrpc.client.Error as err:
            print(f"Problem reloading group: {err}")
            return 1

        return 0

    def add_group(self, arg):
        """Add a process group"""

        if len(arg) == 0:
            print("You need to specify a process group")
            return 1

        name = arg[0]

        try:
            self.server.addgroup(name)
        except xmlrpc.client.Error as err:
            print(f"Problem adding the group: {err}")
            return 1

        return 0

    def remove_group(self, arg):
        """Remove a group"""

        if not arg:
            print("You need to specify a process group")
            return 1

        name = arg[0]

        self.has_group_or_exit(name)

        try:
            self.server.removegroup(arg[0])
        except xmlrpc.client.Error as err:
            print(f"Problem removing the group: {err}")
            return 1

        return 0

    def list(self, arg):
        """List groups or the clients of a group"""

        if len(arg) == 1:
            return self.list_clients(arg[0])

        return self.list_groups()

    def list_clients(self, group):
        """List clients in a group"""

        self.has_group_or_exit(group)

        clients = self.server.listclients(group)
        print(f"There are {len(clients)} clients listed for the group {group}")
        for client, info in clients.items():
            print(client, info)
        return 0

    def list_groups(self):
        """List groups"""

        groups = self.server.listgroups()

        if not groups:
            print("No process groups are registered")
            return 0

        width = max(len(group) for group in groups)

        for group, label in groups.items():
            print(f"{group:{width}} - {label}")

        return 0

    def start_client(self, group, client, args):
        """Start a client"""

        self.has_client_or_exit(group, client)

        try:
            self.server.startclient(group, client, args)
        except xmlrpc.client.Error as err:
            print(f"Problem starting client: {err}")
            return 1

        return 0

    def stop_client(self, group, client):
        """Stop a client"""

        self.has_client_or_exit(group, client)

        try:
            self.server.stopclient(group, client)
        except xmlrpc.client.Error as err:
            print(f"Problem stopping client: {err}")
            return 1

        return 0

    def start_server(self):
        """Start the server daemon"""

        args = [shutil.which("transportd"), "-b"]

        # Long running process, can't use with, so disable warning
        # pylint: disable=consider-using-with

        subprocess.Popen(args)

    def stop_server(self):
        """Stop the server daemon"""

        try:
            self.server.stop()
            time.sleep(1)
        except ConnectionRefusedError:
            pass

        return 0

    def server_control(self, arg):
        """Control server daemon"""

        if len(arg) == 0:
            print("Specify start, stop or restart")
            return 1

        cmd = arg[0]

        if cmd == "start":
            return self.start_server()

        if cmd == "stop":
            return self.stop_server()

        if cmd == "restart":
            self.stop_server()
            return self.start_server()

        print(f"Unknown server command: {cmd}")

        return 1

    def start(self, arg):
        """Start a client or group"""

        if len(arg) == 0:
            print("You need to specify a group or client name ")
            return 1

        if len(arg) > 1:
            if len(arg) == 3:
                args = arg[2]
            else:
                args = ""
            return self.start_client(arg[0], arg[1], args)

        group = arg[0]

        try:
            self.server.startgroup(group)
        except xmlrpc.client.Error as err:
            print(f"Problem starting group: {err}")
            return 1

        print(f"Group started: {group}")

        return 0

    def stop(self, arg):
        """Stop client, group or server"""

        if not arg:
            print("You need to provide a group or client name")
            return 1

        if len(arg) == 2:
            group, client = arg
            return self.stop_client(group, client)

        group = arg[0]

        self.has_group_or_exit(group)

        try:
            self.server.stopgroup(group)
        except xmlrpc.client.Error as err:
            print(f"Porblem stopping group: {err}")
            return 1

        print(f"Group stopped: {group}")

        return 0

    def restart(self, arg):
        """Restart client or group"""

        self.stop(arg)
        self.start(arg)

    def cleanup(self):
        """Stop any transport clients"""

        uid = TransportConfig().uid

        for path in Path("/proc").glob("*"):
            if not path.name.isnumeric():
                continue

            statusfile = path.joinpath("status")

            data = {}

            with statusfile.open("r", encoding="utf=8") as f:
                for line in f.readlines():
                    key, value = line.split(":")
                    data[key] = value.strip()

            if uid != int(data["Uid"].split()[0]):
                continue

            pid = int(data["Pid"])

            print(f"Stopping {pid}: ", end=" ")
            try:
                os.kill(pid, 9)
                print("OK")
            except OSError as err:
                print(f"Error: {err}")

        print("Finished")

        return 0


def main():
    """Script entry point"""

    try:
        control = TransportControl()
        sys.exit(control.run_command(sys.argv))
    except KeyboardInterrupt:
        print("*** BREAK ***")
        sys.exit(1)



#!/usr/bin/env python
"""Create test data files"""

############################################################################
#
#   CreateData
#
#   This script creates test data files.
#
#   History:
#
#   2002-12-12  Todd Valentic
#               Initial implementation
#
#   2022-11-04  Todd Valentic
#               Update for Python3
#
############################################################################

import random
import sys

from datatransport import ProcessClient
from datatransport.utilities import make_path


class CreateData(ProcessClient):
    """Data Source"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

    def init(self):
        super().init()

        self.rate = self.config.get_rate("rate", 60)
        self.path = self.config.get_path("output.path")
        self.names = self.config.get_list("output.names")
        self.maxbytes = self.config.get_bytes("output.maxsize", "10KB")

        if not self.names:
            self.abort("No output names (output.names)")

    def main(self):
        """Main loop"""

        while self.wait(self.rate):

            for name in self.names:

                filename = self.path / name

                make_path(filename)

                try:
                    numbytes = int(random.random() * self.maxbytes)
                    payload = "." * numbytes
                    filename.write_text(payload)
                    self.log.info("Created file: %s (%d bytes)", filename, numbytes)
                except: # pylint: disable=bare-except
                    self.log.exception("Problem creating file")


if __name__ == "__main__":
    CreateData(sys.argv).run()



#!/usr/bin/env python
"""Create parent directories along a path, setting mode if needed"""

#####################################################################
#
#   Create parent directories along a path, setting mode if needed
#
#   2001-12-02  Todd Valentic
#		        Explicity call chmod to fix some issues with umask
#
#   2022-10-12  Todd Valentic
#               Python3 port
#               Use pathlib
#               No longer set group (better done with group setid)
#               Use snake case
#
#   SPDX-License-Identifier: GPL-3.0-or-later
#   Copyright (C) 1999-2022 Todd Valentic
#
#####################################################################

from typing import Union
from pathlib import Path

def make_path(filename: Union[str, Path], mode: int=None) -> None:
    """Create parent directories"""

    for path in reversed(Path(filename).absolute().parents):
        if not path.exists():
            path.mkdir()
            if mode:
                path.chmod(mode)



#!/usr/bin/env python3
"""List newsgroups"""

#####################################################################
#
#   List news groups on a news server
#
#   2012-09-24  Todd Valentic
#               Initial implementation
#
#   2022-10-07  Todd Valentic
#               Python3 updates
#                   optparse -> argparse
#
#   2023-08-17  Todd Valentic
#               Add port option
#
#####################################################################

import argparse
import nntplib


def lister(args):
    """Newsgroup lister"""

    server = nntplib.NNTP(args.newsserver, port=args.port)
    groups = server.list()[1]

    counts = {}

    for group in groups:
        name = group[0]
        first = int(group[2])
        last = int(group[1])
        counts[name] = last - first + 1

    if args.sort == "name":
        key = None
    else:
        key = counts.__getitem__

    for group in sorted(counts, key=key, reverse=args.reverse):
        line = ""
        if args.showcount:
            line += f"{counts[group]:10d} "
        line += group

        print(line)


def main():
    """Script entry point"""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-n",
        "--newsserver",
        default="127.0.0.1",
        help="News server address (default is %(default)s)",
    )

    parser.add_argument(
        "-p", "--port", default=119, type=int, help="New server port (default 119)"
    )

    parser.add_argument(
        "-c", "--showcount", action="store_true", help="Show message counts"
    )

    parser.add_argument(
        "-s", "--sort", default="name", choices=["name", "count"], help="Sort"
    )

    parser.add_argument(
        "-r", "--reverse", action="store_true", help="Reverse sort order"
    )

    args = parser.parse_args()

    lister(args)

    return 0



#!/usr/bin/env python
"""Directory Connect"""

##########################################################################
#
#   XMLRPC Directory Connect
#
#   Connect to the XMLRPC directory service. It will block on startup
#   until connection to the service is established.
#
#   2005-11-09  Todd Valentic
#               Initial implementation.
#
#   2006-07-07  Todd Valentic
#               Added wait at connect. This ensures services
#                   are running when needed.
#
#   2006-09-16  Todd Valentic
#               Added nowait option for foriegn services.
#
#   2006-10-20  Todd Valentic
#               Only print out connected message if waiting.
#
#   2007-04-14  Todd Valentic
#               Set allow_none in connect().
#
#   2009-03-17  Todd Valentic
#               Added async connection option.
#
#   2020-03-09  Todd Valentic
#               Python3: xmlrpc.client
# 		        Change async to defer (async is now a keyword in Python3)
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#               Default to allow None and builtin types
#               No longer a mixin
#
##########################################################################

import xmlrpc.client

from .utilities import xmlrpcdeferred


class Directory:
    """Add methods to connect to the XMLRPC directory service"""

    def __init__(self, parent):
        self.parent = parent

        url = parent.config.get("directory.url")
        kwargs = {"allow_none": True, "use_builtin_types": True}
        self.directory = xmlrpc.client.ServerProxy(url, **kwargs)

        # Wait for the directory service to become ready

        self._waiton(self.directory, "directory")

    def _waiton(self, service, label):
        """Wait until the service is ready"""

        waiting = False
        parent = self.parent

        while parent.is_running():
            try:
                service.ident()
                if waiting:
                    parent.log.info(f"Service '{label}' is ready")
                break
            except:  # pylint: disable=bare-except
                waiting = True
                parent.log.info(f"Waiting for service '{label}' to start")
                parent.wait(2)

    def connect(self, service, defer=False, **kw):
        """Connect to an XMLRPC service via the directory"""

        args = {"allow_none": True, "use_builtin_types": True}

        if defer:
            args["transport"] = xmlrpcdeferred.AsyncTransport()

        args.update(kw)

        url = self.directory.get(service, "url")
        server = xmlrpc.client.ServerProxy(url, **args)
        hold = self.directory.get(service, "hold")

        if hold and not defer:
            self._waiton(server, service)
        else:
            self.parent.log.info(f"Assuming service {service} is ready")

        return server

    def get(self, service, option):
        """Proxy directory service get method"""
        return self.directory.get(service, option)

    def __call__(self, *pos, **kw):
        return self.connect(*pos, **kw)



#!/usr/bin/env python

###################################################################
#
#   Asynchronous XML-RPC client
#
#   This code is a modified version of that posted by A.M. Kichling
#   in the pygtk mailing list [1]. While his was using GTK's async
#   IO handling, I needed something based on asyncore. The changes
#   are pretty small.
#
#   The basic idea is that we provide a custom transport object
#   to the XML-RPC client interface. When the XML-RPC call is 
#   made, instead of reading back the results in a blocking socket,
#   we schedule a "deferred" object. These objects are similar
#   to what Twisted provides, and in this case, they are asyncore
#   dispatcher objects. The main polling loop runs until all of
#   scheduled objects have finished. You can call asyncore.loop()
#   and block or you can have it return periodically if you need
#   to do other work in the meantime.
#
#   The result of the XML-RPC call is accessed through the value
#   property. If the call is complete, it will return the results.
#   If an exception was detected, it will be raised instead.
#
#   When searching for approaches to this problem, I found reference
#   to updating xmlrpc's transport with the more modern
#   HTTPConnection [2]. I haven't implmented this yet. Do we need
#   to do it?
#
#   [1] http://www.mail-archive.com/pygtk@daa.com.au/msg12971.html
#   [2] http://itkovian.net/base/transport-class-pythons-xml-rpc-lib
#
#   2009-03-17  Todd Valentic
#               Initial implementation.
#
#   2020-10-09  Todd Valentic
#               Python3: xmlrpc.client
#
###################################################################

import asyncore
import xmlrpc.client
import socket

class XMLRPCDeferred(asyncore.dispatcher):
    """Object representing the delayed result of an XML-RPC request.

    .is_ready: bool
      True when the result is received; False before then.
    .value : any
      Once is_ready=True, this attribute contains the result of the
      request. If this value is an instance of the xmlrpc.client.Fault
      class, then some exception occurred during the request's
      processing. You can use the getValue() method to extract
      the value and raise an exception if a Fault is found.
      
    """
    
    def __init__ (self, transport, http):
        asyncore.dispatcher.__init__(self,http._conn.sock) 

        self.transport = transport
        self.http = http
        self._value = None
        self.is_ready = False

    def writable (self):
        return False

    def readable (self):
        return not self.is_ready

    def handle_close (self):
        self.close()

    def handle_read (self):

        # We assume that the entire reply is ready.
        # Switch back to blocking so we can pick up all the parts.

        self.socket.setblocking(1)
        errcode, errmsg, headers = self.http.getreply()

        if errcode != 200:
            raise ProtocolError(
                host + handler,
                errcode, errmsg,
                headers
                )

        try:
            result = xmlrpc.client.Transport._parse_response(
                        self.transport,
                        self.http.getfile(), 
                        None)[0]
        except xmlrpc.client.Fault(exc):
            result = exc
            
        self.__value = result
        self.is_ready = True
        self.close()

    def __len__ (self):
        # XXX egregious hack!!!
        # The code in xmlrpc.client.ServerProxy calls len() on the object
        # returned by the transport, and if it's of length 1 returns
        # the contained object.  Therefore, this __len__ method
        # returns a completely fake length of 2.
        return 2 

    def get_value (self):
        # if the value is an XML-RPC fault, raise an exception.

        if not self.is_ready:
            raise IOError('Result not ready')

        if isinstance(self.__value,xmlrpc.client.Fault):
            raise xmlrpc.client.Fault(self.__value)

        return self.__value

    def set_value (self,value):
        self.__value = value

    value = property(get_value,set_value)
    
        
class AsyncTransport (xmlrpc.client.Transport):

    def request(self, host, handler, request_body, verbose=0):
        # issue XML-RPC request

        h = self.make_connection(host)
        if verbose:
            h.set_debuglevel(1)

        self.send_request(h, handler, request_body)
        self.send_host(h, host)
        self.send_user_agent(h)
        self.send_content(h, request_body)

        self.verbose = verbose

        return XMLRPCDeferred(self, h)


if __name__ == '__main__':

    hosts = ['firewall']

    for udu in range(1,9):
        hosts.append('udu%d.face1' % udu)
    
    results = []

    for host in hosts:

        url = 'http://%s.amisr.net:8200' % host
        t = AsyncTransport()
        s = xmlrpc.client.ServerProxy(url,t)
        results.append(s.network_scan())

    print('queries are launched, waiting to finish...')

    asyncore.loop()

    hosts = {}

    for result in results:
        hosts.update(result.value)

    for host in sorted(hosts):
        try:
            hostname = socket.gethostbyaddr(host)[0].replace('.amisr.net','')
        except:
            hostname = ''
        print('[%s]  %-16s  %s' % (hosts[host],host,hostname))





#####################################################################
#
#   PlotTool
#
#   The PlotTool creates plots from incoming data streams using the
#   Round Robin Database (RRD) system. Each time a new record is
#   posted to the monitored newsgroup, the values will be appended
#   to the database and a new plot created. The incoming data format
#   defaults to the ConfigParser (section,key/value) style. However,
#   hooks are provided for subsclassed objects to override this
#   behaviour.
#
#   History:
#
#   1.0.0   2002-10-30  Todd Valentic
#           Initial implementation.
#
#   1.0.1   2003-06-04  Todd Valentic
#           Added time zone info
#           Updated some plot formatting
#
#   1.0.2   2003-06-30  Todd Valentic
#           Updated to new NewsPoller interface
#
#   1.0.3   2003-07-18  Todd Valentic
#           Generalized plotting program.
#
#   1.0.4   2003-08-08  Todd Valentic
#           Use ConfigMixin
#
#   1.0.5   2003-08-26  Todd Valentic
#           Use rrdtool vs RRDtool.
#
#   1.0.6   2003-08-27  Todd Valentic
#           Added min/max checks for variables.
#
#   1.0.7   2004-01-14  Todd Valentic
#           Added ability to set the data set type (DST) and
#               consolidation function (CF) for variables.
#
#   1.0.8   2004-02-15  Todd Valentic
#           Added ability to set start and end time.
#           Added automatic labeling and colors if more then
#               one variable on a plot.
#
#   1.0.9   2004-02-17  Todd Valentic
#           Fixed bug in setting variable range limits. Now
#               explicitly test for None.
#
#   1.0.10  2005-03-17  Todd Valentic
#           Converted from DateTime -> datetime
#           Use ConfigComponent
#           Reworked time zone. Specifically set TZ
#               before updates and plotting.
#           Removed TimeZone objects.
#
#   1.0.11  2006-02-02  Todd Valentic
#           Create output path if necessary.
#
#   1.0.12  2006-07-12 Todd Valentic
#           RRD module changed name again rrdtool -> rrdtoolmodule.
#           Try importing rrdtool then rrdtoolmodule.
#
#   1.0.13  2007-09-18  Todd Valentic
#           Change the file extension on the plots to ".png"
#               I didn't notice that the plot type had changed
#               with RRD 1.2 from .gif!
#
#   1.0.14  2007-10-17  Todd Valentic
#           Fixed bug when a CF other than AVERAGE was used.
#
#   1.0.15  2008-08-11  Todd Valentic
#           Better log info.
#
#   1.0.16  2008-08-20  Todd Valentic
#           Added y.rigid option.
#
#   1.0.17  2008-08-26  Todd Valentic
#           Added CDEF option for variables.
#
#   1.0.18  2008-11-13  Todd Valentic
#           Use SafeConfigParser
#
#   1.0.19  2008-12-08  Todd Valentic
#           Allow line.width to be a float
#
#   1.0.20  2008-12-30  Todd Valentic
#           Only print labels if actually specified.
#           Added logarithmic option to plot
#
#   1.0.21  2009-01-09  Todd Valentic
#           Added misc parameter for plots to add anything
#               other parameters to the graph.
#           Added variable plot style of none to supress any
#               lines or areas from being drawn.
#           Added stack parameter for variables
#
#   1.0.22  2009-01-30  Todd Valentic
#           Tune default line colors and add second (green) area color.
#           Use getComponents()
#           Use NewsTool.asConfig()
#           Make sure ignore_errors is honored
#
#   1.0.23  2009-02-09  Todd Valentic
#           Added plotgroups
#
#   1.0.24  2009-04-28  Todd Valentic
#           Expand plotgroups to be a full component
#           Add matching rules
#           Check for set type (compatibility for python 2.3)
#
#   1.0.25  2009-09-02  Todd Valentic
#           Add filter option for variables to filter out non-numeric
#               characters.
#
#   1.0.26  2009-11-09  Todd Valentic
#           Move plotrate into period.
#
#   1.0.27  2010-02-23  Todd Valentic
#           Do not draw line when stacking area plots
#
#   1.0.28  2010-06-01  Todd Valentic
#           Add debug log entries in check_headers
#
#   1.0.29  2010-09-08  Todd Valentic
#           Allow plotgroup specific variables
#           Add area stack.offset and stack.line
#           Auto increment stack.offset with stack.offset.auto
#           Use get_list()
#           Add area.alpha parameter
#           Use linecolor+alpha for areacolor if it isn't specified.
#           Add header checks per variable
#
#   1.0.30  2011-08-28  Todd Valentic
#           Added hidden variables.
#           Added shadow variables.
#           Added hash for Variable
#
#   1.0.31  2012-05-20  Todd Valentic
#           Added sql support
#               There is a problem with libdbi on CentOS 5. The
#               drivers were not linked properly and don't load.
#               To work around, shell out and run rrdtool. This
#               also gets around running transport in a virtual
#               environment.
#           Added GRID and MGRID colors
#           Added TEXTALIGN
#
#   1.0.32  2012-11-08  Todd Valentic
#           Move plot config file loading to load_plots(). This will
#               all for injecting values into the config file from
#               subclassed versions of PlotTool (example usage
#               would be to pull information from a database). It
#               can also be used to dynamically reload the plot
#               definitions.
#
#   1.0.33  2014-02-16  Todd Valentic
#           Add plot name to localvars when loading variables.
#
#   1.0.34  2015-08-29  Todd Valentic
#           Make sure lookup stack handles plotgroup keys
#           Add font options
#
#   1.0.35  2018-01-16  Todd Valentic
#           Update default line colors (values from matplotlib tab10)
#
#   2022-10-07  Todd Valentic
#               Reorder imports
#               Replace datefunc calls with new methods
#               Use config get_ methods
#               NewsPoller
#
#   2025-04-15  Todd Valentic
#               Escape \ when creating command strings
#
#####################################################################

import fnmatch
import os
import string
import sys
import time

from datetime import datetime
from subprocess import Popen, PIPE

from datatransport import ProcessClient
from datatransport import newstool
from datatransport import NewsPoller
from datatransport import ConfigComponent

# Fake the rrdtool module


class RRDTool:
    def __getattr__(self, name):
        def method(*args):
            args = list(args)
            args.insert(0, name)
            args.insert(0, "rrdtool")
            p = Popen(args, stdout=PIPE, stderr=PIPE)
            out, err = p.communicate()
            if p.returncode != 0:
                raise IOError(err)
            return out, err

        return method


rrdtool = RRDTool()


class ManageTZ:
    def __init__(self):

        if "TZ" in os.environ:
            self.tz = os.environ["TZ"]
        else:
            self.tz = None

    def set(self, newtz):

        os.environ["TZ"] = newtz
        time.tzset()

    def restore(self):

        if self.tz:
            self.set(self.tz)
        else:
            del os.environ["TZ"]
            time.tzset()


timezone = ManageTZ()


def get_headers(self, key):

    headers = {}

    for entry in self.get_list(key, token="\n"):
        if not entry:
            continue
        try:
            header, value = [x.strip() for x in entry.split("=", 1)]
        except:
            self.log.exception("Problem parsing: %s" % entry)
            raise
        headers[header] = value

    return headers


ConfigComponent.get_headers = get_headers

####################################################################
#
#   Period
#
#   This component represents a time period over which to plot
#   data. The values are:
#
#   span - rrd time description for how long to plot from the
#          end of the archive.
#
#   xlabel - String to print as the x-axis label
#   suffix - Appended to filename (before extension). Defaults to the
#            value of span. The current time is applied in strftime
#            style.
#   ticks  - Formatting string for the x-tick labels (rrd style)
#
#   Examples:
#       span;   1day
#       xlabel: %A, %B, %d, %Y
#       suffix: 1day
#       ticks:  MINUTE:30:HOUR:2:HOUR:2:0:%H:%M
#
#   Plots are generated at intervals of plotrate, which can be
#   independently set for each period.
#
####################################################################


class Period(ConfigComponent):
    def __init__(self, name, config):
        ConfigComponent.__init__(self, "period", name, config)

        self.span = self.get("span", "1hour")
        self.xlabel = self.get("xlabel", "")
        self.suffix = self.get("suffix", name)
        self.ticks = self.get("ticks", None)
        self.plotrate = self.get_timedelta("plotrate", 60)

        self.log.info("      period: %s" % name)
        self.log.info("        span    : %s" % self.span)
        self.log.info("        xlabel  : %s" % self.xlabel)
        self.log.info("        suffix  : %s" % self.suffix)
        self.log.info("        ticks   : %s" % self.ticks)
        self.log.info("        plotrate: %s" % self.plotrate)

        self.plotrate = self.plotrate.total_seconds()
        self.next_plot_time = time.time()

    def ready_to_plot(self):
        return time.time() > self.next_plot_time

    def schedule_next_plot(self):
        waittime = self.plotrate - time.time() % self.plotrate
        self.next_plot_time = time.time() + waittime


####################################################################
#
#   Variable
#
#   This component represents a variable to store in the database
#   and draw on the plot. It consists of a section/key pair that
#   is used to look up the value in the ConfigParser-style input
#   file. By default, the key is just the name.
#
#   We assume that the timestamps of the variable are given in UTC.
#   If they are not, you need to manipulate the timestamp in a
#   derived class.
#
#   If the style is none, the database reference is included,
#   but no line or area is plotted (the intention being that
#   a CDEF in plot is going to use these variables to generate
#   it's own line).
#
#   style: area | line | none
#
####################################################################


class Variable(ConfigComponent):
    def __init__(self, name, parent, **kw):
        ConfigComponent.__init__(self, "variable", name, parent, localvars=kw)

        self.plotgroup = kw["plotgroup"]

        self.lookup_stack.append("variable.%s.%s." % (self.plotgroup, name))

        if self.plotgroup and not os.path.exists(self.plotgroup):
            os.makedirs(self.plotgroup)

        self.sql = self.get("sql")
        self.sqlfunc = self.get("sql.func", "avg")
        self.key = self.get("key", name)
        self.section = self.get("section", "DEFAULT")
        self.label = self.get("label", "")
        self.linecolor = self.get("line.color", "")
        self.linewidth = self.get_float("line.width", 1)
        self.areacolor = self.get("area.color")
        self.areaalpha = self.get("area.alpha")
        self.style = self.get("style", "line").lower()
        self.min_value = self.get_float("min", None)
        self.max_value = self.get_float("max", None)
        self.datatype = self.get("datatype", "GAUGE")
        self.cf = self.get("function", "AVERAGE")
        self.cdef = self.get("cdef", None)
        self.stack = self.get("stack", False)
        self.stackoffset = self.get("stack.offset")
        self.stackauto = self.get_boolean("stack.offset.auto", False)
        self.stackline = self.get_boolean("stack.line", False)
        self.filter = self.get_boolean("filter", False)
        self.headers = self.get_headers("headers")
        self.start = self.get_datetime("start", "")
        self.end = self.get_datetime("end", "")
        self.shadow = self.get("shadow")

        self.log.info("      variable: %s" % name)
        self.log.info("        plotgroup   :  %s" % self.plotgroup)
        self.log.info("        shadow      :  %s" % self.shadow)
        self.log.info("        key         :  %s" % self.key)
        self.log.info("        section     :  %s" % self.section)
        self.log.info("        label       :  %s" % self.label)
        self.log.info("        style       :  %s" % self.style)
        self.log.info("        line.color  :  %s" % self.linecolor)
        self.log.info("        line.width  :  %s" % self.linewidth)
        self.log.info("        area.color  :  %s" % self.areacolor)
        self.log.info("        range.min   :  %s" % self.min_value)
        self.log.info("        range.max   :  %s" % self.max_value)
        self.log.info("        data type   :  %s" % self.datatype)
        self.log.info("        function    :  %s" % self.cf)
        self.log.info("        cdef        :  %s" % self.cdef)
        self.log.info("        stack       :  %s" % self.stack)
        self.log.info("        stack.offset:  %s" % self.stackoffset)
        self.log.info("        filter      :  %s" % self.filter)
        self.log.info("        headers     :  %s" % self.headers)
        self.log.info("        start       :  %s" % self.start)
        self.log.info("        end         :  %s" % self.end)
        self.log.info("        sql         :  %s" % self.sql)

        if self.start:
            self.start = ":start=%s" % self.start.strftime(r"%H\:%M %Y%m%d")

        if self.end and not self.start:
            self.start = r":start=00\:00 01/01/1980"

        if self.end:
            self.end = ":end=%s" % self.end.strftime(r"%H\:%M %Y%m%d")

        if self.shadow:
            self.database = os.path.join(self.plotgroup, self.shadow + ".rrd")
        elif self.sql:
            self.database = self.sql
        else:
            self.database = os.path.join(self.plotgroup, name + ".rrd")
            if not os.path.exists(self.database):
                self.create_database()

        self.linecolors = [
            "2B337F",
            "ff7f0e",
            "2ca02c",
            "d62728",
            "9467bd",
            "8c564b",
            "e377c2",
            "e377c2",
            "7f7f7f",
            "17becf",
        ]

        self.areacolors = ["AAAADD", "66AF6C"]

    def __hash__(self):
        return hash("%s:%s" % (self.name, self.plotgroup))

    def __eq__(self, other):
        return other.__hash__() == self.__hash__()

    def __repr__(self):
        return "%s:%s" % (self.name, self.plotgroup)

    def create_database(self):

        oneday = 60 * 60 * 24

        samplerate = self.get_timedelta("samplerate", 60)
        timeout = self.get_timedelta("timeout", 60)
        duration = self.get_timedelta("duration", oneday)

        samplerate = samplerate.total_seconds()
        timeout = timeout.total_seconds()
        duration = duration.total_seconds()

        startTime = time.time() - duration

        cmd = []
        cmd.append(self.database)
        cmd.append("-b %d" % startTime)
        cmd.append("-s %d" % samplerate)
        cmd.append("RRA:%s:0.5:1:%d" % (self.cf, int(duration / samplerate)))

        if self.min_value is not None:
            minvalue = str(self.min_value)
        else:
            minvalue = "U"

        if self.max_value is not None:
            maxvalue = str(self.max_value)
        else:
            maxvalue = "U"

        cmd.append(
            "DS:%s:%s:%d:%s:%s"
            % (self.name, self.datatype, timeout, minvalue, maxvalue)
        )

        try:
            rrdtool.create(*tuple(cmd))
        except:
            timezone.restore()
            self.log.exception("Error creating database")
            self.abort("Exiting")

        self.log.info("Created initial database for %s" % self.name)

    def update(self, timestamp, value):

        if self.shadow or self.sql:
            return

        entry = "%d" % timestamp.timestamp()

        if value:
            entry += ":" + str(value)
        else:
            entry += ":U"

        # self.log.debug('%s updated at %s' % (self.name,timestamp))
        rrdtool.update(self.database, entry)

    def options(self, index, hidden=False):

        params = []
        line = "LINE%f" % self.linewidth
        name = self.name

        if self.shadow:
            dsname = self.shadow
        elif self.sql:
            dsname = self.sqlfunc
        else:
            dsname = self.name

        params.append(
            "DEF:%s=%s:%s:%s%s%s"
            % (self.name, self.database, dsname, self.cf, self.start, self.end)
        )

        if self.label:
            label = self.label
        else:
            label = ""

        if self.linecolor:
            linecolor = self.linecolor
        else:
            linecolor = self.linecolors[index % len(self.linecolors)]

        if self.areacolor:
            areacolor = self.areacolor
        else:
            areacolor = linecolor

        if self.areaalpha and len(areacolor) == 6:
            areacolor += self.areaalpha

        if self.cdef:
            name = self.name + "cdef"
            params.append("CDEF:%s=%s" % (name, self.cdef))

        if self.stack:
            stack = ":STACK"
        else:
            stack = ""

        if self.stackauto and self.stackoffset:
            offset = float(self.stackoffset) * index
        else:
            offset = self.stackoffset

        if hidden:
            style = "none"
        else:
            style = self.style

        if style == "area":

            offsetline = "%s:%s#%s:" % (line, offset, linecolor)

            if self.stackoffset:
                params.append(offsetline)

            params.append("AREA:%s#%s:%s%s" % (name, areacolor, label, stack))

            if self.stack:
                if self.stackline:
                    if self.stackoffset:
                        params.append(offsetline)
                    params.append("%s:%s#%s::STACK" % (line, name, linecolor))
            else:
                params.append("%s:%s#%s:" % (line, name, linecolor))

        elif style == "line":
            params.append("%s:%s#%s:%s%s" % (line, name, linecolor, label, stack))

        return params


####################################################################
#
#   Plot
#
#   This component represents a plot. It includes a list of
#   variable components and time periods. For each period, the
#   a graph of the variables is created. The other parameters
#   decorate the graph (and will be the same for each plot).
#
####################################################################


class Plot(ConfigComponent):
    def __init__(self, plotname, parent, **kw):
        ConfigComponent.__init__(self, "plot", plotname, parent, localvars=kw)

        self.log.info("    plot: %s" % plotname)

        self.plotgroup = kw["plotgroup"]

        groupkey = "plot.%s.%s." % (self.plotgroup, plotname)
        self.lookup_stack.append(groupkey)

        for mixin in self.get_list(groupkey + "mixin")[::-1]:
            self.lookup_stack.insert(-1, mixin + ".")

        self.log.debug("    Extended lookup stack: %s" % self.lookup_stack)

        self.variables = []
        self.periods = []
        self.hidden_variables = []

        varlocal = {}
        varlocal.update(kw)
        varlocal["plot"] = plotname

        for name in self.get_list("variables", ""):
            self.variables.append(Variable(name, parent, **varlocal))

        for name in self.get_list("variables.hidden", ""):
            self.hidden_variables.append(Variable(name, parent, **varlocal))

        for name in self.get_list("periods", ""):
            self.periods.append(Period(name, parent))

        if not self.periods:
            self.log.warn("No periods defined for %s" % plotname)

        self.filename = self.get("filename", plotname)
        self.title = self.get("title", None)
        self.ymin = self.get_float("y.min", None)
        self.ymax = self.get_float("y.max", None)
        self.ylabel = self.get("y.label", None)
        self.yticks = self.get("y.ticks", None)
        self.yrigid = self.get_boolean("y.rigid", False)
        self.height = self.get_int("height", None)
        self.width = self.get_int("width", None)
        self.exponent = self.get_int("exponent", None)
        self.legend = self.get_int("legend", 1)
        self.griddash = self.get("grid.dash", "1:0")
        self.shadea = self.get("color.shadea", "FFFFFF")
        self.shadeb = self.get("color.shadeb", "FFFFFF")
        self.background = self.get("color.background", "FFFFFF")
        self.gridcolor = self.get("color.grid", "00000033")
        self.mgridcolor = self.get("color.mgrid", "00000033")
        self.textalign = self.get("textalign", "left")
        self.tz = self.get("timezone")
        self.timestart = self.get("time.start", "now")
        self.timestop = self.get("time.end", "now")
        self.logarithmic = self.get_boolean("logarithmic", False)
        self.misc = self.get("misc")

        self.log.info("options:")

        for key, value in self.optionsdict().items():
            self.log.info("  %s: %s" % (key, value))

        outputpath = os.path.dirname(self.filename)

        if not outputpath:
            outputpath = self.plotgroup
            self.filename = os.path.join(outputpath, self.filename)

        if not os.path.exists(outputpath):
            try:
                os.makedirs(outputpath)
            except:
                self.log.error("Problem creating output path: %s" % outputpath)
                raise

    def options(self, period):

        options = []

        options.append("%s-%s.png" % (self.filename, period.suffix))
        options.append("--color=SHADEA#%s" % self.shadea)
        options.append("--color=SHADEB#%s" % self.shadeb)
        options.append("--color=BACK#%s" % self.background)
        options.append("--color=GRID#%s" % self.gridcolor)
        options.append("--color=MGRID#%s" % self.mgridcolor)
        options.append("TEXTALIGN:%s" % self.textalign)

        options.append("-s %s-%s" % (self.timestart, period.span))
        options.append("-e %s" % (self.timestop))

        if period.xlabel:
            xlabel = time.strftime(period.xlabel, time.localtime())
            options.append(r"COMMENT:%s\c" % xlabel)

        if self.yticks:
            options.append("-y %s" % self.yticks)

        if self.logarithmic:
            options.append("--logarithmic")

        if not self.legend:
            options.append("-g")

        if not self.ylabel is None:
            options.append("-v %s" % self.ylabel)

        if not self.title is None:
            options.append("-t %s" % self.title)

        if not self.height is None:
            options.append("-h %d" % self.height)

        if not self.width is None:
            options.append("-w %d" % self.width)

        if not self.ymin is None:
            options.append("-l %f" % self.ymin)

        if not self.ymax is None:
            options.append("-u %f" % self.ymax)

        if self.yrigid:
            options.append("-r")

        if self.griddash:
            options.append("--grid-dash=%s" % self.griddash)

        if not self.exponent is None:
            options.append("--units-exponent=%d" % self.exponent)

        if period.ticks:
            options.append("--x-grid=%s" % period.ticks)

        for index, variable in enumerate(self.variables):
            options.extend(variable.options(index))

        for index, variable in enumerate(self.hidden_variables):
            options.extend(variable.options(index, hidden=True))

        if self.misc:
            options.extend(self.misc.split("\n"))

        return tuple(options)

    def create(self):

        if self.tz:
            timezone.set(self.tz)

        for period in self.periods:

            if not period.ready_to_plot():
                continue

            if self.parent.parent.is_stopped():
                break

            try:
                options = self.options(period)
                self.log.info(
                    "Plotting %s %s at %s" % (self.plotgroup, self.name, period.name)
                )
                self.log.debug("  options: %s" % options)
                rrdtool.graph(*options)
            except:
                self.log.exception(
                    "Problem creating plots for %s %s" % (self.name, period.name)
                )

            period.schedule_next_plot()

        if self.tz:
            timezone.restore()


####################################################################
#
#   PlotGroup
#
#   Orginizational structure for producing a group of plots.
#
####################################################################


class PlotGroup(ConfigComponent):
    def __init__(self, name, parent):
        kw = {"plotgroup": name}
        ConfigComponent.__init__(self, "plotgroup", name, parent, kw)

        self.log.info("  Plot group %s" % name)

        self.headers = self.get_headers("headers")

        if self.headers:
            self.log.info("    headers:")
            for header, value in self.headers.items():
                self.log.info("      %s: %s" % (header, value))

        self.log.info("    local vars:")

        self.localvars = {"plotgroup": name}

        for key, option in self.optionsdict().items():
            value = self.get(option, raw=True)
            self.localvars["plotgroup." + key] = value
            self.log.info("      %s: %s" % (key, value))

        self.plots = self.get_components("plots", Plot, **self.localvars).items()


####################################################################
#
#   PlotTool
#
#   Process client to create plots. The database will be updated
#   each time a new record shows up in the monitored news group,
#   Independently, the plots will updated at a periodic rate. If
#   no new records are received, the plots will continue to update.
#
#   It will be common that clients will inherit from this class
#   in order to interpret the message in the news group and make
#   it look like a ConfigParser object.
#
####################################################################


class PlotTool(ProcessClient):
    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.news_poller = NewsPoller(
            self, callback=self.process, 
            idleFunc=self.create_plots
        )
        self.main = self.news_poller.main

        self.ignore_errors = self.get_boolean("ignore_errors", False)
        self.report_missing = self.get_boolean("report_missing", True)

        self.load_plots()

        if not self.plots:
            self.abort("No plots were defined")

        self.log.info("Variable set:")
        for variable in self.variables:
            self.log.info("  %s" % variable)

        # When filtering, keep digits and +-.

        self.filter_table = string.maketrans(string.printable, string.printable)
        self.filter_chars = string.letters + string.punctuation
        self.filter_chars = self.filter_chars.translate(self.filter_table, "+-.")

    def load_plots(self):

        self.log.info("Loading plot group definitions:")

        self.plotgroups = self.get_components("plotgroups", PlotGroup)

        if not self.plotgroups:
            self.plotgroups = {"": PlotGroup("", self)}

        self.headers = {}
        self.plots = []
        self.variables = set()

        for plotgroup in self.plotgroups.values():
            self.plots.extend(plotgroup.plots)
            for plot in plotgroup.plots:
                self.variables.update(plot.variables)
                self.variables.update(plot.hidden_variables)

    def create_plots(self):

        for plot in self.plots:
            plot.create()
            if self.is_stopped():
                break

    def check_headers(self, variable):

        plotgroup = self.plotgroups[variable.plotgroup]

        self.log.debug("Checking headers:")
        self.log.debug("  plotgroup.headers: %s" % plotgroup.headers)
        self.log.debug("  variable.headers:  %s" % variable.headers)
        self.log.debug("  message  headers:  %s" % self.headers)

        target_headers = {}
        target_headers.update(plotgroup.headers)
        target_headers.update(variable.headers)

        for header, pattern in target_headers.items():
            if header in self.headers:
                if not fnmatch.fnmatch(self.headers[header], pattern):
                    self.log.debug("  reject: header mismatch (%s)" % header)
                    return False
            else:
                self.log.debug("  reject: header missing (%s)" % header)
                return False

        self.log.debug("  accept: all headers match")
        return True

    def update_dateabase(self):

        self.log.info("Updating variables for %s" % self.timestamp)

        for variable in self.variables:

            if not self.check_headers(variable):
                continue

            label = "%s %s" % (variable.plotgroup, variable.name)

            try:
                value = self.get_value(variable)
                if variable.filter:
                    value = value.translate(self.filter_table, self.filter_chars)
            except:
                if self.ignore_errors:
                    if self.report_missing:
                        self.log.exception("No data found for %s" % label)
                    continue
                else:
                    self.log.error("Problem with %s" % label)
                    raise

            self.log.info("  %s = %s" % (label, value))

            try:
                variable.update(self.timestamp, value)
            except:
                self.log.exception("Problem updating %s" % label)
                if not self.ignore_errors:
                    raise

    def get_value(self, variable):
        return self.data.get(variable.section, variable.key)

    def get_timestamp(self, message):
        return newstool.message_date(message)

    def parse(self, message):

        try:
            self.data = newstool.as_config(message)
        except:
            self.log.exception("Problem parsing input")
            return False

        return True

    def process(self, message):

        # Need self.data, self.timestamp and self.headers

        if self.parse(message):
            self.timestamp = self.get_timestamp(message)
            self.headers = dict(message.items())
            self.update_dateabase()


def main():
    PlotTool(sys.argv).run()



#!/usr/bin/env python3
"""Text posting example"""

##########################################################################
#
#   Posting example
#
#   Periodically post a message to a news group
#
#   2022-10-10  Todd Valentic
#               Initial implementation
#
##########################################################################

import sys

from datatransport import ProcessClient
from datatransport import NewsPoster


class Client(ProcessClient):
    """Process Client"""

    def __init__(self, argv):
        ProcessClient.__init__(self, argv)

        self.news_poster = NewsPoster(self)

        self.rate = self.config.get_rate("rate")
        self.text = self.config.get("text")

        self.log.info("Post message every %s", self.rate.period)

    def main(self):
        """Main application"""

        while self.wait(self.rate):
            self.news_poster.post_text(self.text)
            self.log.info("Posted message")


if __name__ == "__main__":
    Client(sys.argv).run()



#!/usr/bin/env python

from Transport 		import ProcessClient
from Transport		import NewsPollMixin

import sys

class Reader(ProcessClient,NewsPollMixin):
	
	def __init__(self,argv):
		ProcessClient.__init__(self,argv)
		NewsPollMixin.__init__(self,callback=self.process)

	def process(self,message):
		
		body = message.get_payload()

		self.log.info('Message: %s' % body)

if __name__ == '__main__':
	Reader(sys.argv).run()
	



#!/usr/bin/env python3
"""Make Files"""

##########################################################################
#
#   Make test files
#
#   2023-08-13  Todd Valentic
#               Initial implementation
#
##########################################################################

import os
import sys

from pathlib import Path

from datatransport import ProcessClient
from datatransport import ConfigComponent
from datatransport.utilities import size_desc

class DataFile(ConfigComponent):

    def __init__(self, *p, **kw):
        ConfigComponent.__init__(self, 'datafile', *p, **kw)

        self.output_size = self.config.get_bytes('size', '1kb')
        self.output_name = self.config.get_path('filename', 'data.bin') 

    def process(self):
        """Create data file"""

        filename = Path(self.now().strftime(str(self.output_name)))

        filename.parent.mkdir(parents=True, exist_ok=True)

        with filename.open('wb') as output:
            output.write(os.urandom(self.output_size))

        self.log.info("Wrote %s (%s)", filename, size_desc(self.output_size))

class FileMaker(ProcessClient):
    
    def __init__(self, args):
        ProcessClient.__init__(self, args)

        self.datafiles = self.config.get_components('datafiles', factory=DataFile)  
        self.rate = self.config.get_rate('rate')

    def main(self):
        """Main application"""

        while self.wait(self.rate):
            
            for datafile in self.datafiles.values():
                datafile.process()


if __name__ == '__main__':
    FileMaker(sys.argv).run()




#!/usr/bin/env python

###########################################################
#
#   Simulated genset data producer
#
#   Write out data records to be picked up by another
#   process.
#
#   2010-08-21  Todd Valentic
#               Initial implementation.
#
###########################################################

from Transport      import ProcessClient

import sys
import random
import ConfigParser
import StringIO

class Genset(ProcessClient):

    def __init__(self,argv):
        ProcessClient.__init__(self,argv)

        self.outputPath = self.get('output.path')
        self.rate       = self.getRate('rate')

    def run(self):

        while self.wait(self.rate):
            try:
                self.process()
            except:
                self.log.exception('Problem processing')

    def process(self):

        data = self.createData()

        output = self.currentTime().strftime(self.outputPath)

        open(output,'w').write(data)

        self.log.info('Created: %s' % output)

    def createData(self):

        data = ConfigParser.ConfigParser()

        data.set('DEFAULT','timestamp',str(self.currentTime()))

        data.add_section('E1')
        data.add_section('E2')
        data.add_section('E3')

        data.set('E1','temperature',str(random.uniform(60,90)))
        data.set('E2','temperature',str(random.uniform(60,90)))
        data.set('E3','temperature',str(random.uniform(60,90)))

        buffer = StringIO.StringIO()
        data.write(buffer)

        return buffer.getvalue()

if __name__ == '__main__':
    Genset(sys.argv).run()

